# +
import warnings

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import ListedColormap
from matplotlib.gridspec import GridSpec

def html_to_rgb(html_color):
    r"""
    Convert HTML hex color code to RGB tuple.
    
    Arguments:
        html_color: HTML hex color string (e.g., '#FF0000' or 'FF0000')
        
    Returns:
        rgb_color: RGB tuple with values 0-255
    """
    # 去掉颜色代码前的 `#`
    html_color = html_color.lstrip('#')

    # 处理简写形式的颜色代码（如 `#RGB` 或 `#RGBA`）
    if len(html_color) == 3:
        html_color = ''.join([c*2 for c in html_color])
    elif len(html_color) == 4:
        html_color = ''.join([c*2 for c in html_color])

    # 只保留前六位颜色信息，忽略透明度
    if len(html_color) == 6 or len(html_color) == 8:
        html_color = html_color[:6]
    else:
        raise ValueError("Invalid HTML color code length")

    # 将十六进制颜色代码转换为 RGB 元组
    rgb_color = tuple(int(html_color[i:i+2], 16) for i in (0, 2, 4))
    return rgb_color

def get_rgb_function(cmap, min_value, max_value):
    r"""
    Generate a function to map continuous values to RGB using colormap.
    
    Arguments:
        cmap: Matplotlib colormap object
        min_value: Minimum value for color mapping
        max_value: Maximum value for color mapping
        
    Returns:
        func: Function that maps values to RGB colors
    """
    r"""Generate a function to map continous values to RGB values using colormap between min_value & max_value."""

    if min_value > max_value:
        raise ValueError("Max_value should be greater or than min_value.")

    if min_value == max_value:
        warnings.warn(
            "Max_color is equal to min_color. It might be because of the data or bad parameter choice. "
            "If you are using plot_contours function try increasing max_color_quantile parameter and"
            "removing cell types with all zero values."
        )

        def func_equal(x):
            factor = 0 if max_value == 0 else 0.5
            return cmap(np.ones_like(x) * factor)

        return func_equal

    def func(x):
        return cmap((np.clip(x, min_value, max_value) - min_value) / (max_value - min_value))

    return func


def rgb_to_ryb(rgb):
    r"""
    Convert colors from RGB colorspace to RYB (red-yellow-blue) colorspace.
    
    Arguments:
        rgb: RGB color array with shape (N, 3) or (3,)
        
    Returns:
        ryb: RYB color array with same shape as input
    """
    rgb = np.array(rgb)
    if len(rgb.shape) == 1:
        rgb = rgb[np.newaxis, :]

    white = rgb.min(axis=1)
    black = (1 - rgb).min(axis=1)
    rgb = rgb - white[:, np.newaxis]

    yellow = rgb[:, :2].min(axis=1)
    ryb = np.zeros_like(rgb)
    ryb[:, 0] = rgb[:, 0] - yellow
    ryb[:, 1] = (yellow + rgb[:, 1]) / 2
    ryb[:, 2] = (rgb[:, 2] + rgb[:, 1] - yellow) / 2

    mask = ~(ryb == 0).all(axis=1)
    if mask.any():
        norm = ryb[mask].max(axis=1) / rgb[mask].max(axis=1)
        ryb[mask] = ryb[mask] / norm[:, np.newaxis]

    return ryb + black[:, np.newaxis]


def ryb_to_rgb(ryb):
    r"""
    Convert colors from RYB (red-yellow-blue) colorspace to RGB colorspace.
    
    Arguments:
        ryb: RYB color array with shape (N, 3) or (3,)
        
    Returns:
        rgb: RGB color array with same shape as input
    """
    ryb = np.array(ryb)
    if len(ryb.shape) == 1:
        ryb = ryb[np.newaxis, :]

    black = ryb.min(axis=1)
    white = (1 - ryb).min(axis=1)
    ryb = ryb - black[:, np.newaxis]

    green = ryb[:, 1:].min(axis=1)
    rgb = np.zeros_like(ryb)
    rgb[:, 0] = ryb[:, 0] + ryb[:, 1] - green
    rgb[:, 1] = green + ryb[:, 1]
    rgb[:, 2] = (ryb[:, 2] - green) * 2

    mask = ~(ryb == 0).all(axis=1)
    if mask.any():
        norm = rgb[mask].max(axis=1) / ryb[mask].max(axis=1)
        rgb[mask] = rgb[mask] / norm[:, np.newaxis]

    return rgb + white[:, np.newaxis]




def plot_spatial_general(
        adata,
    value_df,
    coords,
    labels,
    text=None,
    circle_diameter=4.0,
    alpha_scaling=1.0,
    max_col=(np.inf, np.inf, np.inf, np.inf, np.inf, np.inf, np.inf),
    max_color_quantile=0.98,
    show_img=True,
    img=None,
    img_alpha=1.0,
    adjust_text=False,
    plt_axis="off",
    axis_y_flipped=True,
    x_y_labels=("", ""),
    crop_x=None,
    crop_y=None,
    text_box_alpha=0.9,
    reorder_cmap=range(7),
    style="fast",
    colorbar_position="bottom",
    colorbar_label_kw={},
    colorbar_shape={},
    colorbar_tick_size=12,
    colorbar_grid=None,
    image_cmap="Greys_r",
    white_spacing=20,
    palette=None,
    legend_title_fontsize=12,
    return_ax=False,
    
):
    r"""
    Create spatial plot with color gradient and interpolation for cell type abundances.
    
    Supports up to 7 cell types with default colors: yellow, orange, blue, green, purple, grey, white.
    
    Arguments:
        adata: Annotated data object
        value_df: DataFrame with cell abundances or features (max 7 columns) across locations
        coords: Array with x,y coordinates for plotting spots
        labels: List of cell type labels
        text: DataFrame with x,y coordinates and text for annotations (None)
        circle_diameter: Diameter of spot circles (4.0)
        alpha_scaling: Color transparency adjustment factor (1.0)
        max_col: Maximum colorscale values for each column ((np.inf, np.inf, np.inf, np.inf, np.inf, np.inf, np.inf))
        max_color_quantile: Quantile threshold for colorscale cropping (0.98)
        show_img: Whether to display background image (True)
        img: Background tissue image array (None)
        img_alpha: Transparency of background image (1.0)
        adjust_text: Whether to adjust text labels to prevent overlap (False)
        plt_axis: Axis display setting ('off')
        axis_y_flipped: Whether to flip y-axis to match image coordinates (True)
        x_y_labels: Axis labels as (x_label, y_label) (('', ''))
        crop_x: X-axis cropping limits as (min, max) (None)
        crop_y: Y-axis cropping limits as (min, max) (None)
        text_box_alpha: Transparency of text boxes (0.9)
        reorder_cmap: Color order indices for categories (range(7))
        style: Plot style - 'fast' or 'dark_background' ('fast')
        colorbar_position: Colorbar position - 'bottom', 'right', or None ('bottom')
        colorbar_label_kw: Keyword arguments for colorbar labels ({})
        colorbar_shape: Colorbar shape parameters ({})
        colorbar_tick_size: Colorbar tick label size (12)
        colorbar_grid: Colorbar grid dimensions as (rows, cols) (None, auto-determined)
        image_cmap: Colormap for grayscale background image ('Greys_r')
        white_spacing: Percentage of colorbar hidden as white space (20)
        palette: Custom color palette as list or dict (None, uses defaults)
        legend_title_fontsize: Font size for legend titles (12)
        return_ax: Whether to return axes object (False)
        
    Returns:
        fig: matplotlib.figure.Figure object, or (fig, ax) if return_ax=True
    """

    if value_df.shape[1] > 7:
        raise ValueError("Maximum of 7 cell types / factors can be plotted at the moment")

    def create_colormap(R, G, B):
        spacing = int(white_spacing * 2.55)

        N = 255
        M = 3

        alphas = np.concatenate([[0] * spacing * M, np.linspace(0, 1.0, (N - spacing) * M)])

        vals = np.ones((N * M, 4))
        #         vals[:, 0] = np.linspace(1, R / 255, N * M)
        #         vals[:, 1] = np.linspace(1, G / 255, N * M)
        #         vals[:, 2] = np.linspace(1, B / 255, N * M)
        for i, color in enumerate([R, G, B]):
            vals[:, i] = color / 255
        vals[:, 3] = alphas

        return ListedColormap(vals)
    if palette==None:
        # Create linearly scaled colormaps
        YellowCM = create_colormap(240, 228, 66)  # #F0E442 ['#F0E442', '#D55E00', '#56B4E9',
        # '#009E73', '#5A14A5', '#C8C8C8', '#323232']
        RedCM = create_colormap(213, 94, 0)  # #D55E00
        BlueCM = create_colormap(86, 180, 233)  # #56B4E9
        GreenCM = create_colormap(0, 158, 115)  # #009E73
        GreyCM = create_colormap(200, 200, 200)  # #C8C8C8
        WhiteCM = create_colormap(50, 50, 50)  # #323232
        PurpleCM = create_colormap(90, 20, 165)  # #5A14A5

        cmaps = [YellowCM, RedCM, BlueCM, GreenCM, PurpleCM, GreyCM, WhiteCM]

        cmaps = [cmaps[i] for i in reorder_cmap]
    else:
        if isinstance(palette, list):
            cmaps = [html_to_rgb(i) for i in palette]
            cmaps = [create_colormap(*i) for i in cmaps]
        elif isinstance(palette, dict):
            #adata uns to color dict
            cmaps = [html_to_rgb(palette[i]) for i in value_df.columns]
            cmaps = [create_colormap(*i) for i in cmaps]
        else:
            raise ValueError("palette should be a list or dict or None")


    with mpl.style.context(style):
        fig = plt.figure()

        if colorbar_position == "right":
            if colorbar_grid is None:
                colorbar_grid = (len(labels), 1)

            shape = {"vertical_gaps": 1.5, "horizontal_gaps": 0, "width": 0.15, "height": 0.2}
            shape = {**shape, **colorbar_shape}

            gs = GridSpec(
                nrows=colorbar_grid[0] + 2,
                ncols=colorbar_grid[1] + 1,
                width_ratios=[1, *[shape["width"]] * colorbar_grid[1]],
                height_ratios=[1, *[shape["height"]] * colorbar_grid[0], 1],
                hspace=shape["vertical_gaps"],
                wspace=shape["horizontal_gaps"],
            )
            ax = fig.add_subplot(gs[:, 0], aspect="equal", rasterized=True)

        if colorbar_position == "bottom":
            if colorbar_grid is None:
                if len(labels) <= 3:
                    colorbar_grid = (1, len(labels))
                else:
                    n_rows = round(len(labels) / 3 + 0.5 - 1e-9)
                    colorbar_grid = (n_rows, 3)

            shape = {"vertical_gaps": 0.3, "horizontal_gaps": 0.6, "width": 0.2, "height": 0.035}
            shape = {**shape, **colorbar_shape}

            gs = GridSpec(
                nrows=colorbar_grid[0] + 1,
                ncols=colorbar_grid[1] + 2,
                width_ratios=[0.3, *[shape["width"]] * colorbar_grid[1], 0.3],
                height_ratios=[1, *[shape["height"]] * colorbar_grid[0]],
                hspace=shape["vertical_gaps"],
                wspace=shape["horizontal_gaps"],
            )

            ax = fig.add_subplot(gs[0, :], aspect="equal", rasterized=True)

        if colorbar_position is None:
            ax = fig.add_subplot(aspect="equal", rasterized=True)

        if colorbar_position is not None:
            cbar_axes = []
            for row in range(1, colorbar_grid[0] + 1):
                for column in range(1, colorbar_grid[1] + 1):
                    cbar_axes.append(fig.add_subplot(gs[row, column]))

            n_excess = colorbar_grid[0] * colorbar_grid[1] - len(labels)
            if n_excess > 0:
                for i in range(1, n_excess + 1):
                    cbar_axes[-i].set_visible(False)

        ax.set_xlabel(x_y_labels[0])
        ax.set_ylabel(x_y_labels[1])

        if img is not None and show_img:
            ax.imshow(img, aspect="equal", alpha=img_alpha, origin="lower", cmap=image_cmap)

        # crop images in needed
        if crop_x is not None:
            ax.set_xlim(crop_x[0], crop_x[1])
        if crop_y is not None:
            ax.set_ylim(crop_y[0], crop_y[1])

        if axis_y_flipped:
            ax.invert_yaxis()

        if plt_axis == "off":
            for spine in ax.spines.values():
                spine.set_visible(False)
            ax.tick_params(bottom=False, labelbottom=False, left=False, labelleft=False)

        counts = value_df.values.copy()

        # plot spots as circles
        c_ord = list(np.arange(0, counts.shape[1]))

        colors = np.zeros((*counts.shape, 4))
        weights = np.zeros(counts.shape)

        for c in c_ord:
            min_color_intensity = counts[:, c].min()
            max_color_intensity = np.min([np.quantile(counts[:, c], max_color_quantile), max_col[c]])

            rgb_function = get_rgb_function(cmap=cmaps[c], min_value=min_color_intensity, max_value=max_color_intensity)

            color = rgb_function(counts[:, c])
            color[:, 3] = color[:, 3] * alpha_scaling

            norm = mpl.colors.Normalize(vmin=min_color_intensity, vmax=max_color_intensity)

            if colorbar_position is not None:
                cbar_ticks = [
                    min_color_intensity,
                    np.mean([min_color_intensity, max_color_intensity]),
                    max_color_intensity,
                ]
                cbar_ticks = np.array(cbar_ticks)

                if max_color_intensity > 13:
                    cbar_ticks = cbar_ticks.astype(np.int32)
                else:
                    cbar_ticks = cbar_ticks.round(2)

                cbar = fig.colorbar(
                    mpl.cm.ScalarMappable(norm=norm, cmap=cmaps[c]),
                    cax=cbar_axes[c],
                    orientation="horizontal",
                    extend="both",
                    ticks=cbar_ticks,
                )

                cbar.ax.tick_params(labelsize=colorbar_tick_size)
                max_color = rgb_function(max_color_intensity / 1.5)
                cbar.ax.set_title(labels[c], **{**{"size": legend_title_fontsize, "color": max_color, "alpha": 1}, **colorbar_label_kw})

            colors[:, c] = color
            weights[:, c] = np.clip(counts[:, c] / (max_color_intensity + 1e-10), 0, 1)
            weights[:, c][counts[:, c] < min_color_intensity] = 0

        colors_ryb = np.zeros((*weights.shape, 3))

        for i in range(colors.shape[0]):
            colors_ryb[i] = rgb_to_ryb(colors[i, :, :3])

        def kernel(w):
            return w**2

        kernel_weights = kernel(weights[:, :, np.newaxis])
        weighted_colors_ryb = (colors_ryb * kernel_weights).sum(axis=1) / kernel_weights.sum(axis=1)

        weighted_colors = np.zeros((weights.shape[0], 4))

        weighted_colors[:, :3] = ryb_to_rgb(weighted_colors_ryb)

        weighted_colors[:, 3] = colors[:, :, 3].max(axis=1)

        ax.scatter(x=coords[:, 0], y=coords[:, 1], c=weighted_colors, s=circle_diameter**2)

        # add text
        if text is not None:
            bbox_props = dict(boxstyle="round", ec="0.5", alpha=text_box_alpha, fc="w")
            texts = []
            for x, y, s in zip(
                np.array(text.iloc[:, 0].values).flatten(),
                np.array(text.iloc[:, 1].values).flatten(),
                text.iloc[:, 2].tolist(),
            ):
                texts.append(ax.text(x, y, s, ha="center", va="bottom", bbox=bbox_props))

            if adjust_text:
                from adjustText import adjust_text

                adjust_text(texts, arrowprops=dict(arrowstyle="->", color="w", lw=0.5))
    if return_ax==True:
        return fig,ax
    else:
        return fig


def plot_spatial(adata, color, img_key="hires", show_img=True, **kwargs):
    r"""
    Create spatial plot from Visium data with color gradient and interpolation.
    
    Supports up to 7 cell types with default colors: yellow, orange, blue, green, purple, grey, white.
    
    Arguments:
        adata: AnnData object with spatial coordinates in adata.obsm['spatial']
        color: List of column names from adata.obs to plot
        img_key: Image resolution key - 'hires' or 'lowres' ('hires')
        show_img: Whether to display background tissue image (True)
        **kwargs: Additional arguments passed to plot_spatial_general
        
    Returns:
        fig: matplotlib.figure.Figure object
    """

    if show_img is True:
        kwargs["show_img"] = True
        kwargs["img"] = list(adata.uns["spatial"].values())[0]["images"][img_key]

    # location coordinates
    if "spatial" in adata.uns.keys():
        kwargs["coords"] = (
            adata.obsm["spatial"] * list(adata.uns["spatial"].values())[0]["scalefactors"][f"tissue_{img_key}_scalef"]
        )
    else:
        kwargs["coords"] = adata.obsm["spatial"]

    fig = plot_spatial_general(adata,value_df=adata.obs[color], **kwargs)  # cell abundance values
    fig.axes[0].set_xlim(kwargs["coords"][:,0].min(),kwargs["coords"][:,0].max())
    fig.axes[0].set_ylim(kwargs["coords"][:,1].max(),kwargs["coords"][:,1].min())

    return fig

def create_colormap(R, G, B):
    r"""
    Create a matplotlib colormap from RGB values with alpha gradient.
    
    Arguments:
        R: Red component (0-255)
        G: Green component (0-255)
        B: Blue component (0-255)
        
    Returns:
        colormap: matplotlib.colors.ListedColormap object
    """
    spacing = int(20 * 2.55)

    N = 255
    M = 3

    alphas = np.concatenate([[0] * spacing * M, np.linspace(0, 1.0, (N - spacing) * M)])

    vals = np.ones((N * M, 4))
    #         vals[:, 0] = np.linspace(1, R / 255, N * M)
    #         vals[:, 1] = np.linspace(1, G / 255, N * M)
    #         vals[:, 2] = np.linspace(1, B / 255, N * M)
    for i, color in enumerate([R, G, B]):
        vals[:, i] = color / 255
    vals[:, 3] = alphas

    return ListedColormap(vals)

def spatial_value(adata,color,library_id,
            dot_size=4,white_spacing=20,
            colorbar_show=True,
            colorbar_tick_size=12,cmap=create_colormap(255, 0, 0),
            legend_title_fontsize=12,
            legend_title_color=None,
            colorbar_label_kw={},res='hires',
            img_show=True,ax=None,alpha_img=1):
    r"""
    Plot spatial expression values for a single gene or feature on tissue image.
    
    Arguments:
        adata: AnnData object with spatial data
        color: Gene name or obs column to visualize
        library_id: Spatial library identifier
        dot_size: Size of spots on the plot (4)
        white_spacing: White space percentage in colormap (20)
        colorbar_show: Whether to display colorbar (True)
        colorbar_tick_size: Font size for colorbar ticks (12)
        cmap: Colormap for values (create_colormap(255, 0, 0))
        legend_title_fontsize: Font size for legend title (12)
        legend_title_color: Color for legend title (None, auto-determined)
        colorbar_label_kw: Additional colorbar label arguments ({})
        res: Image resolution - 'hires' or 'lowres' ('hires')
        img_show: Whether to show background tissue image (True)
        ax: Existing matplotlib axes object (None)
        alpha_img: Transparency of background image (1)
        
    Returns:
        ax: matplotlib.axes.Axes object
    """

    #ax_input=False
    if ax is None:
        fig, ax = plt.subplots(figsize=(4,4))

    img=adata.uns["spatial"][library_id]["images"][res]
    if img_show:
        ax.imshow(img, aspect="equal", alpha=alpha_img, origin="lower", cmap=cmap)
        ax.invert_yaxis()
    ax.axis(False)
    coords=adata.obsm['spatial']*adata.uns['spatial'][library_id]['scalefactors'][f'tissue_{res}_scalef']
    if color in adata.obs.columns:
        weighted_colors=adata.obs[color].values.reshape(-1)
    elif color in adata.var_names:
        weighted_colors=adata[:,color].to_df().values.reshape(-1)
    elif adata.raw is not None and color in adata.raw.var_names:
        weighted_colors=adata.raw[:,color].to_adata().to_df().values.reshape(-1)
    ax.scatter(x=coords[:, 0], y=coords[:, 1], c=weighted_colors, s=dot_size**2,
              cmap=cmap)
    
    ax.set_xlim(coords[:,0].min(),coords[:,0].max())
    ax.set_ylim(coords[:,1].max(),coords[:,1].min())
    
    
    colorbar_grid=None
    labels=[color]
    colorbar_shape={}
    if colorbar_grid is None:
        if len(labels) <= 3:
            colorbar_grid = (1, len(labels))
        else:
            n_rows = round(len(labels) / 3 + 0.5 - 1e-9)
            colorbar_grid = (n_rows, 3)
    
    shape = {"vertical_gaps": 1, "horizontal_gaps": 0.6, "width": 0.5, "height": 0.05}
    shape = {**shape, **colorbar_shape}
    
    gs = GridSpec(
        nrows=colorbar_grid[0] + 1,
        ncols=colorbar_grid[1] + 2,
        width_ratios=[0.3, *[shape["width"]] * colorbar_grid[1], 0.3],
        height_ratios=[1, *[shape["height"]] * colorbar_grid[0]],
        hspace=shape["vertical_gaps"],
        wspace=shape["horizontal_gaps"],
    )
    
    #ax1 = fig.add_subplot(gs[0, :], aspect="equal", rasterized=True)
    if colorbar_show is True:
        fig=plt.gcf()
        cbar_axes = []
        for row in range(1, colorbar_grid[0] + 1):
            for column in range(1, colorbar_grid[1] + 1):
                cbar_axes.append(fig.add_subplot(gs[row, column]))
        
        n_excess = colorbar_grid[0] * colorbar_grid[1] - len(labels)
        if n_excess > 0:
            for i in range(1, n_excess + 1):
                cbar_axes[-i].set_visible(False)
        
        norm=mpl.colors.Normalize(vmin=min(weighted_colors), vmax=max(weighted_colors))
        rgb_function = get_rgb_function(cmap=cmap,
                                        min_value=min(weighted_colors), max_value=max(weighted_colors))

        cbar = fig.colorbar(
            mpl.cm.ScalarMappable(norm=norm, cmap=cmap),
            cax=cbar_axes[0],
            orientation="horizontal",
            extend="both",
            #ticks=cbar_ticks,
        )
        cbar.ax.tick_params(labelsize=colorbar_tick_size)
        if legend_title_color is None:
            max_color = rgb_function(max(weighted_colors) / 1.5)
        else:
            max_color=legend_title_color
        cbar.ax.set_title(labels[0], **{**{"size": legend_title_fontsize, 
                                        "color": max_color, "alpha": 1},
                                            **colorbar_label_kw})

    return ax


import numpy as np
import pandas as pd
import torch
import torch.nn.functional as F
from torch.utils.data import Dataset

from ._starfysh import LOGGER


#---------------------------
# Single Sample dataloader
#---------------------------

class VisiumDataset(Dataset):
    """
    Loading a single preprocessed ST AnnData, gene signature & Anchor spots for Starfysh training
    """

    def __init__(
        self,
        adata,
        args,
    ):
        spots = adata.obs_names
        genes = adata.var_names

        x = adata.X if isinstance(adata.X, np.ndarray) else adata.X.A
        self.expr_mat = pd.DataFrame(x, index=spots, columns=genes)
        self.gexp = args.sig_mean_norm
        self.anchor_idx = args.pure_idx
        self.library_n = args.win_loglib

    def __len__(self):
        return len(self.expr_mat)

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()
        sample = torch.Tensor(
            np.array(self.expr_mat.iloc[idx, :], dtype='float')
        )

        return (sample,
                torch.Tensor(self.gexp.iloc[idx, :]),  # normalized signature exprs
                torch.Tensor(self.anchor_idx[idx, :]),  # anchors
                torch.Tensor(self.library_n[idx,None]),  # library size
               )


class VisiumPoEDataSet(VisiumDataset):
    
    def __init__(
        self,
        adata,
        args,
    ):
        super(VisiumPoEDataSet, self).__init__(adata, args)
        self.image = args.img.astype(np.float64)
        self.map_info = args.map_info
        self.r = args.params['patch_r']
        self.spot_img_stack = []

        self.density_std = args.img.std()

        assert self.image is not None,\
            "Empty paired H&E image," \
            "please use regular `Starfysh` without PoE integration" \
            "if your dataset doesn't contain histology image"

        # Retrieve image patch around each spot
        scalef = args.scalefactor['tissue_hires_scalef']  # High-res scale factor
        h, w = self.image.shape[:2]
        patch_dim = (self.r*2, self.r*2, 3) if self.image.ndim == 3 else (self.r*2, self.r*2)

        for i in range(len(self.expr_mat)):
            xc = int(np.round(self.map_info.iloc[i]['imagecol'] * scalef))
            yc = int(np.round(self.map_info.iloc[i]['imagerow'] * scalef))

            # boundary conditions: edge spots
            yl, yr = max(0, yc-self.r), min(self.image.shape[0], yc+self.r)
            xl, xr = max(0, xc-self.r), min(self.image.shape[1], xc+self.r)
            top = max(0, self.r-yc)
            bottom = h if h > (yc+self.r) else h-(yc+self.r)
            left = max(0, self.r-xc)
            right = w if w > (xc+self.r) else w-(xc+self.r)

            #try:
            patch = np.zeros(patch_dim)
            patch[top:bottom, left:right] = self.image[yl:yr, xl:xr]
            self.spot_img_stack.append(patch)
            #except ValueError:
            #    LOGGER.warning('Skipping the patch loading of an edge spot...')


    def __len__(self):
        return len(self.expr_mat)

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()
        sample = torch.Tensor(
            np.array(self.expr_mat.iloc[idx, :], dtype='float')
        )
        spot_img_stack = self.spot_img_stack[idx]
        return (sample,
                torch.Tensor(self.anchor_idx[idx, :]),
                torch.Tensor(self.library_n[idx, None]),
                spot_img_stack,
                self.map_info.index[idx],
                torch.Tensor(self.gexp.iloc[idx, :]),
               )
    
#---------------------------
# Integrative Dataloader
#--------------------------- 


class IntegrativeDataset(VisiumDataset):
    """
    Loading multiple preprocessed ST sample AnnDatas, gene signature & Anchor spots for Starfysh training
    """

    def __init__(
        self,
        adata,
        args,
    ):
        super(IntegrativeDataset, self).__init__(adata, args)
        self.image = args.img
        self.map_info = args.map_info
        self.r = args.params['patch_r']

    def __len__(self):
        return len(self.expr_mat)

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()
        sample = torch.Tensor(
            np.array(self.expr_mat.iloc[idx, :], dtype='float')
        )
        return (sample,   
                torch.Tensor(self.gexp.iloc[idx, :]),
                torch.Tensor(self.anchor_idx[idx, :]),
                torch.Tensor(self.library_n[idx, None])
               )
    
    
class IntegrativePoEDataset(VisiumDataset):

    def __init__(
        self,
        adata,
        args,
    ):
        super(IntegrativePoEDataset, self).__init__(adata, args)
        self.image = args.img
        self.map_info = args.map_info
        self.r = args.params['patch_r']
        

        assert self.image is not None,\
            "Empty paired H&E image," \
            "please use regular `Starfysh` without PoE integration" \
            "if your dataset doesn't contain histology image"
        spot_img_all = []
        
        # Retrieve image patch around each spot
        for sample_id in args.img.keys():
            
            scalef_i = args.scalefactor[sample_id]['tissue_hires_scalef']  # High-res scale factor
            h, w = self.image[sample_id].shape[:2]
            patch_dim = (self.r*2, self.r*2, 3) if self.image[sample_id].ndim == 3 else (self.r*2, self.r*2)

            list_ = adata.obs['sample'] == sample_id
            for i in range(len(self.expr_mat.loc[list_,:])):
                xc = int(np.round(self.map_info.loc[list_,:].iloc[i]['imagecol'] * scalef_i))
                yc = int(np.round(self.map_info.loc[list_,:].iloc[i]['imagerow'] * scalef_i))

                # boundary conditions: edge spots
                yl, yr = max(0, yc-self.r), min(self.image[sample_id].shape[0], yc+self.r)
                xl, xr = max(0, xc-self.r), min(self.image[sample_id].shape[1], xc+self.r)
                top = max(0, self.r-yc)
                bottom = h if h > (yc+self.r) else h-(yc+self.r)
                left = max(0, self.r-xc)
                right = w if w > (xc+self.r) else w-(xc+self.r)
            
                try:
                    patch = np.zeros(patch_dim)
                    patch[top:bottom, left:right] = self.image[sample_id][yl:yr, xl:xr]
                    spot_img_all.append(patch)
                except ValueError:
                    LOGGER.warning('Skipping the patch loading of an edge spot...')                
                
        self.spot_img_stack = list(spot_img_all)
        #print(self.spot_img_stack.shape)

    def __len__(self):
        return len(self.expr_mat)

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()
        sample = torch.Tensor(
            np.array(self.expr_mat.iloc[idx, :], dtype='float')
        )
        return (sample,   
                torch.Tensor(self.anchor_idx[idx, :]),
                torch.Tensor(self.library_n[idx, None]),
                self.spot_img_stack[idx],
                self.map_info.index[idx],
                torch.Tensor(self.gexp.iloc[idx, :]),
               )
    



r"""
NoCD evaluation metrics for community detection.

This module provides comprehensive evaluation metrics for assessing the quality
of community detection results in single-cell data. It includes both supervised
metrics (when ground truth is available) and unsupervised metrics (intrinsic
quality measures).

Supervised metrics:
    - Accuracy-based measures for comparing with ground truth
    - Classification metrics (precision, recall, F1-score)
    - Information-theoretic measures (mutual information, entropy)
    - Clustering-specific metrics (ARI, silhouette score)

Unsupervised metrics:
    - Modularity and quality measures
    - Silhouette analysis
    - Calinski-Harabasz index
    - Davies-Bouldin index
    - Internal consistency measures

Features:
- Designed specifically for biological data characteristics
- Handles high-dimensional sparse data
- Supports both discrete and continuous evaluation
- Integrates with standard scikit-learn metrics
"""

from .supervised import *
from .unsupervised import *



import numpy as np
import pandas as pd

import scanpy as sc
from scipy.stats import rankdata
from sklearn.metrics import pairwise_distances
from tqdm import tqdm

from . import core


def prepare_multiome_anndata(
    atac_ad, rna_ad, SEACells_label="SEACell", n_bins_for_gc=50
):
    """Function to create metacell Anndata objects from single-cell Anndata objects for multiome data.

    :param atac_ad: (Anndata) ATAC Anndata object with raw peak counts in `X`. These anndata objects should be constructed
     using the example notebook available in
    :param rna_ad: (Anndata) RNA Anndata object with raw gene expression counts in `X`. Note: RNA and ATAC anndata objects
     should contain the same set of cells
    :param SEACells_label: (str) `atac_ad.obs` field for constructing metacell matrices. Same field will be used for
      summarizing RNA and ATAC metacells.
    :param n_bins_gc: (int) Number of bins for creating GC bins of ATAC peaks.
    :return: ATAC metacell Anndata object and RNA metacell Anndata object.
    """
    # Subset of cells common to ATAC and RNA
    common_cells = atac_ad.obs_names.intersection(rna_ad.obs_names)
    if len(common_cells) != atac_ad.shape[0]:
        print(
            "Warning: The number of cells in RNA and ATAC objects are different. Only the common cells will be used."
        )
    atac_mod_ad = atac_ad[common_cells, :]
    rna_mod_ad = rna_ad[common_cells, :]

    # #################################################################################
    # Generate metacell matrices

    # Set of metacells
    metacells = atac_mod_ad.obs[SEACells_label].astype(str).unique()
    metacells = metacells[atac_mod_ad.obs[SEACells_label].value_counts()[metacells] > 1]

    print("Generating Metacell matrices...")
    print(" ATAC")
    atac_meta_ad = core.summarize_by_SEACell(
        atac_mod_ad, SEACells_label=SEACells_label, summarize_layer="X"
    )
    atac_meta_ad = atac_meta_ad[metacells, :]
    # ATAC - Summarize SVD representation

    svd = pd.DataFrame(atac_mod_ad.obsm["X_svd"], index=atac_mod_ad.obs_names)
    summ_svd = svd.groupby(atac_mod_ad.obs[SEACells_label]).mean()
    atac_meta_ad.obsm["X_svd"] = summ_svd.loc[atac_meta_ad.obs_names, :].values

    # ATAC - Normalize
    _add_atac_meta_data(atac_meta_ad, atac_mod_ad, n_bins_for_gc)
    sc.pp.filter_genes(atac_meta_ad, min_cells=1)
    _normalize_ad(atac_meta_ad)

    # RNA summaries using ATAC SEACells
    print(" RNA")
    rna_mod_ad.obs["temp"] = atac_mod_ad.obs[SEACells_label]
    rna_meta_ad = core.summarize_by_SEACell(
        rna_mod_ad, SEACells_label="temp", summarize_layer="X"
    )
    rna_meta_ad = rna_meta_ad[metacells, :]
    _normalize_ad(rna_meta_ad)

    return atac_meta_ad, rna_meta_ad


def _normalize_ad(meta_ad, save_raw=True):
    if save_raw:
        # Save in raw
        meta_ad.raw = meta_ad.copy()

    # Normalize
    sc.pp.normalize_total(meta_ad, key_added="n_counts")
    sc.pp.log1p(meta_ad)


def _add_atac_meta_data(atac_meta_ad, atac_ad, n_bins_for_gc):
    atac_ad.var["log_n_counts"] = np.ravel(np.log10(atac_ad.X.sum(axis=0)))

    atac_meta_ad.var["GC_bin"] = np.digitize(
        atac_ad.var["GC"], np.linspace(0, 1, n_bins_for_gc)
    )
    atac_meta_ad.var["counts_bin"] = np.digitize(
        atac_ad.var["log_n_counts"],
        np.linspace(
            atac_ad.var["log_n_counts"].min(),
            atac_ad.var["log_n_counts"].max(),
            n_bins_for_gc,
        ),
    )


def _pyranges_from_strings(pos_list):
    """Function to create pyranges for a `pd.Series` of strings."""
    # Chromosome and positions
    chr = pos_list.str.split(":").str.get(0)
    start = pd.Series(pos_list.str.split(":").str.get(1)).str.split("-").str.get(0)
    end = pd.Series(pos_list.str.split(":").str.get(1)).str.split("-").str.get(1)

    # Create ranges
    import pyranges as pr
    gr = pr.PyRanges(chromosomes=chr, starts=start, ends=end)
    return gr


def _pyranges_to_strings(peaks):
    """Function to convert pyranges to `pd.Series` of strings of format 'chr:start-end'."""
    # Chromosome and positions
    chr = peaks.Chromosome.astype(str).values
    start = peaks.Start.astype(str).values
    end = peaks.End.astype(str).values

    # Create ranges
    gr = chr + ":" + start + "-" + end

    return gr


def load_transcripts(path_to_gtf):
    """Load transcripts from GTF File. `chr` is preprended to each entry."""
    import pyranges as pr
    gtf = pr.read_gtf(path_to_gtf)
    gtf.Chromosome = "chr" + gtf.Chromosome.astype(str)
    transcripts = gtf[gtf.Feature == "transcript"]
    return transcripts


def _peaks_correlations_per_gene(
    gene,
    atac_exprs,
    rna_exprs,
    atac_meta_ad,
    peaks_pr,
    transcripts,
    span,
    n_rand_sample=100,
):
    # Gene transcript - use the longest transcript
    gene_transcripts = transcripts[transcripts.gene_name == gene]
    if len(gene_transcripts) == 0:
        return 0
    longest_transcript = gene_transcripts[
        np.arange(len(gene_transcripts))
        == np.argmax(gene_transcripts.End - gene_transcripts.Start)
    ]
    start = longest_transcript.Start.values[0] - span
    end = longest_transcript.End.values[0] + span

    # Gene span
    import pyranges as pr
    gene_pr = pr.from_dict(
        {
            "Chromosome": [longest_transcript.Chromosome.values[0]],
            "Start": [start],
            "End": [end],
        }
    )
    gene_peaks = peaks_pr.overlap(gene_pr)
    if len(gene_peaks) == 0:
        return 0
    gene_peaks_str = _pyranges_to_strings(gene_peaks)

    # Compute correlations
    X = atac_exprs.loc[:, gene_peaks_str].T
    cors = 1 - np.ravel(
        pairwise_distances(
            np.apply_along_axis(rankdata, 1, X.values),
            rankdata(rna_exprs[gene].T.values).reshape(1, -1),
            metric="correlation",
        )
    )
    cors = pd.Series(cors, index=gene_peaks_str)

    # Random background
    df = pd.DataFrame(1.0, index=cors.index, columns=["cor", "pval"])
    df["cor"] = cors
    for p in df.index:
        # TODO: Handle exception properly
        try:
            # Try random sampling without replacement
            rand_peaks = np.random.choice(
                atac_meta_ad.var_names[
                    (atac_meta_ad.var["GC_bin"] == atac_meta_ad.var["GC_bin"][p])
                    & (
                        atac_meta_ad.var["counts_bin"]
                        == atac_meta_ad.var["counts_bin"][p]
                    )
                ],
                n_rand_sample,
                False,
            )
        except:  # noqa: E722
            rand_peaks = np.random.choice(
                atac_meta_ad.var_names[
                    (atac_meta_ad.var["GC_bin"] == atac_meta_ad.var["GC_bin"][p])
                    & (
                        atac_meta_ad.var["counts_bin"]
                        == atac_meta_ad.var["counts_bin"][p]
                    )
                ],
                n_rand_sample,
                True,
            )

        if type(atac_exprs) is sc.AnnData:
            X = pd.DataFrame(atac_exprs[:, rand_peaks].X.todense().T)
        else:
            X = atac_exprs.loc[:, rand_peaks].T

        rand_cors = 1 - np.ravel(
            pairwise_distances(
                np.apply_along_axis(rankdata, 1, X.values),
                rankdata(rna_exprs[gene].T.values).reshape(1, -1),
                metric="correlation",
            )
        )

        m = np.mean(rand_cors)
        v = np.std(rand_cors)

        from scipy.stats import norm

        df.loc[p, "pval"] = 1 - norm.cdf(cors[p], m, v)

    return df


def get_gene_peak_correlations(
    atac_meta_ad,
    rna_meta_ad,
    path_to_gtf,
    gene_ranges=None,
    span=100000,
    n_jobs=1,
    gene_set=None,
):
    """Function to compute  correlations between gene expression and peak accessibility.

    :param atac_meta_ad: (Anndata) ATAC metacell Anndata created using `prepare_multiome_anndata`
    :param rna_meta_ad: (Anndata) RNA metacell Anndata created using `prepare_multiome_anndata`
    :param path_to_gtf: (str or None) Path to ENSEMBL GTF file OR None if using pyranges object as input
    :param gene_ranges: (pyranges or None) Pyranges object containing regions corresponding to custom annotation sets. Only used if path_to_gtf is None.
    :param span: (int) Genomic window around the gene body to identify for which correlations with expression are computed
    :param n_jobs: (int) Number of jobs for parallel processing
    :param gene_set: (pd.Series) Subset of genes for which to compute correlations. All genes are used by default

    :return: `pd.Series` with a dataframe of correlation and p-value for each gene. Note that p-value is one-sided assuming positive correlations
    """
    # #################################################################################
    print("Loading transcripts per gene...")
    if path_to_gtf is None:
        transcripts = gene_ranges
    else:
        transcripts = load_transcripts(path_to_gtf)

    print("Preparing matrices for gene-peak associations")
    atac_exprs = pd.DataFrame(
        atac_meta_ad.X.todense(),
        index=atac_meta_ad.obs_names,
        columns=atac_meta_ad.var_names,
    )
    rna_exprs = pd.DataFrame(
        rna_meta_ad.X.todense(),
        index=rna_meta_ad.obs_names,
        columns=rna_meta_ad.var_names,
    )
    peaks_pr = _pyranges_from_strings(atac_meta_ad.var_names)

    print("Computing peak-gene correlations")
    if gene_set is None:
        use_genes = rna_meta_ad.var_names
    else:
        use_genes = gene_set
    from joblib import Parallel, delayed

    gene_peak_correlations = Parallel(n_jobs=n_jobs)(
        delayed(_peaks_correlations_per_gene)(
            gene, atac_exprs, rna_exprs, atac_meta_ad, peaks_pr, transcripts, span
        )
        for gene in tqdm(use_genes)
    )
    gene_peak_correlations = pd.Series(gene_peak_correlations, index=use_genes)
    return gene_peak_correlations


def get_gene_peak_assocations(gene_peak_correlations, pval_cutoff=1e-1, cor_cutoff=0.1):
    """Determine the number of significantly correlated peaks per gene.

    :param gene_peak_correlations: (pd.Series) Output of `get_gene_peak_correlations` function
    :param p_val_cutoff: (float) Nominal p-value cutoff for test of significance of correlation
    :param cor_cutoff: (float) Correlation cutoff

    :return: `pd.Series` with number of significantly positive correlated peaks with each gene
    """
    peak_counts = pd.Series(0, index=gene_peak_correlations.index)
    for gene in tqdm(peak_counts.index):
        df = gene_peak_correlations[gene]
        if type(df) is int:
            continue
        gene_peaks = df.index[(df["pval"] < pval_cutoff) & (df["cor"] > cor_cutoff)]
        peak_counts[gene] = len(gene_peaks)

    return peak_counts


def get_gene_scores(
    atac_meta_ad, gene_peak_correlations, pval_cutoff=1e-1, cor_cutoff=0.1
):
    """Compute the aggregate accessibility of all peaks associated with each gene.

    Gene scores are computed as the aggregate accessibility of all the signficantly correlated peaks associated with a gene.

    :param atac_meta_ad: (Anndata) ATAC metacell Anndata created using `prepare_multiome_anndata`
    :param gene_peak_correlations: (pd.Series) Output of `get_gene_peak_correlations` function
    :param p_val_cutoff: (float) Nominal p-value cutoff for test of significance of correlation
    :param cor_cutoff: (float) Correlation cutoff

    :return: `pd.DataFrame` of ATAC gene scores (cells X genes)
    """
    gene_scores = pd.DataFrame(
        0.0, index=atac_meta_ad.obs_names, columns=gene_peak_correlations.index
    )

    for gene in tqdm(gene_scores.columns):
        df = gene_peak_correlations[gene]
        if type(df) is int:
            continue
        gene_peaks = df.index[(df["pval"] < pval_cutoff) & (df["cor"] > cor_cutoff)]
        gene_scores[gene] = np.ravel(
            np.dot(atac_meta_ad[:, gene_peaks].X.todense(), df.loc[gene_peaks, "cor"])
        )
    gene_scores = gene_scores.loc[:, (gene_scores.sum() >= 0)]
    return gene_scores



# 扩展版本：支持所有模式的3D可视化
import numpy as np
import plotly.graph_objects as go
from sklearn.neighbors import NearestNeighbors
from scipy.stats import norm
from scipy.integrate import odeint

# 兼容性导入
try:
    import plotly.colors.qualitative as pq
except ImportError:
    try:
        import plotly.colors as pc
        pq = pc.qualitative
    except ImportError:
        # 如果plotly不可用，提供备用颜色
        class MockColors:
            Plotly = ['#636EFA', '#EF553B', '#00CC96', '#AB63FA', '#FFA15A', '#19D3F3', '#FF6692', '#B6E880', '#FF97FF', '#FECB52']
            Light24 = ['#FD3216', '#00FE35', '#6A76FC', '#FED4C4', '#FE00CE', '#0DF9FF', '#F6F926', '#FF9616', '#479B55', '#EEA6FB']
            Dark24 = ['#2E91E5', '#E15F99', '#1CA71C', '#FB0D0D', '#DA16FF', '#222A2A', '#B68100', '#750D86', '#EB663B', '#511CFB']
            Alphabet = ['#AA0DFE', '#3283FE', '#85660D', '#782AB6', '#565656', '#1C8356', '#16FF32', '#F7E1A0', '#E2E2E2', '#1CBE4F']
        pq = MockColors()

def plot_cell_communication_3D(
    adata,
    database_name: str = None,
    pathway_name: str = None,
    lr_pair = None,
    keys = None,
    plot_method: str = "cell",
    background: str = "summary",
    background_legend: bool = False,
    clustering: str = None,
    summary: str = "sender",
    cmap: str = "coolwarm",
    cluster_cmap: dict = None,
    pos_idx: np.ndarray = np.array([0, 1], int),  # 现在支持[0,1,2]
    ndsize: float = 1,
    scale: float = 1.0,
    normalize_v: bool = False,
    normalize_v_quantile: float = 0.95,
    arrow_color: str = "#333333",
    grid_density: float = 1.0,
    grid_knn: int = None,
    grid_scale: float = 1.0,
    grid_thresh: float = 1.0,
    grid_width: float = 0.005,
    stream_density: float = 1.0,
    stream_linewidth: float = 1,
    stream_cutoff_perc: float = 5,
    filename: str = None,
    ax = None,
    # 新增3D参数
    grid_cutoff_perc: float = 90,
    html_3d: bool = False,  # 强制生成HTML 3D输出
    camera_eye: dict = None  # 3D相机位置
):
    """
    扩展版本：支持所有模式的3D可视化
    
    新参数:
    html_3d : bool, default False
        如果为True，生成3D HTML而不是matplotlib图
    camera_eye : dict, optional
        3D相机位置，例如 {'x': 1.5, 'y': 1.5, 'z': 1.5}
    """
    
    # 检测是否为3D模式
    is_3d = len(pos_idx) == 3 or html_3d
    
    # 原有数据提取逻辑保持不变
    if not keys is None:
        ncell = adata.shape[0]
        V = np.zeros([ncell, len(pos_idx)], float)
        signal_sum = np.zeros([ncell], float)
        for key in keys:
            if summary == 'sender':
                V_key = adata.obsm['commot_sender_vf-'+database_name+'-'+key]
                # 处理2D->3D扩展
                if len(pos_idx) == 3 and V_key.shape[1] == 2:
                    V_key_extended = np.zeros([V_key.shape[0], 3])
                    V_key_extended[:, :2] = V_key
                    V += V_key_extended[:, pos_idx]
                else:
                    V += V_key[:, pos_idx]
                signal_sum += adata.obsm['commot-'+database_name+"-sum-sender"]['s-'+key]
            elif summary == 'receiver':
                V_key = adata.obsm['commot_receiver_vf-'+database_name+'-'+key]
                if len(pos_idx) == 3 and V_key.shape[1] == 2:
                    V_key_extended = np.zeros([V_key.shape[0], 3])
                    V_key_extended[:, :2] = V_key
                    V += V_key_extended[:, pos_idx]
                else:
                    V += V_key[:, pos_idx]
                signal_sum += adata.obsm['commot-'+database_name+"-sum-receiver"]['r-'+key]
        V = V / float(len(keys))
        signal_sum = signal_sum / float(len(keys))
    else:
        # 单个通路处理
        if not lr_pair is None:
            vf_name = database_name+'-'+lr_pair[0]+'-'+lr_pair[1]
            sum_name = lr_pair[0]+'-'+lr_pair[1]
        elif not pathway_name is None:
            vf_name = database_name+'-'+pathway_name
            sum_name = pathway_name
        else:
            vf_name = database_name+'-total-total'
            sum_name = 'total-total'
        
        if summary == 'sender':
            V_raw = adata.obsm['commot_sender_vf-'+vf_name]
            signal_sum = adata.obsm['commot-'+database_name+"-sum-sender"]['s-'+sum_name]
        elif summary == 'receiver':
            V_raw = adata.obsm['commot_receiver_vf-'+vf_name]
            signal_sum = adata.obsm['commot-'+database_name+"-sum-receiver"]['r-'+sum_name]
        
        # 处理维度扩展
        if len(pos_idx) == 3 and V_raw.shape[1] == 2:
            V = np.zeros([V_raw.shape[0], 3])
            V[:, :2] = V_raw
            V = V[:, pos_idx]
        else:
            V = V_raw[:, pos_idx]
    
    # 获取空间坐标
    X = adata.obsm["spatial"][:, pos_idx]
    
    # 向量预处理
    if normalize_v:
        V = V / np.quantile(np.linalg.norm(V, axis=1), normalize_v_quantile)
    
    # 3D HTML模式 (支持所有方法)
    if is_3d:
        return create_3d_html_plot(
            X, V, signal_sum, adata, plot_method, background, clustering, 
            cmap, cluster_cmap, summary, ndsize, scale,
            grid_density, grid_knn, grid_scale, grid_thresh,
            stream_density, stream_linewidth, stream_cutoff_perc,
            arrow_color, filename, camera_eye, background_legend,
            grid_cutoff_perc
        )
    
    # 原有2D模式 - 调用原来的函数逻辑
    else:
        return plot_original_2d(
            X, V, signal_sum, adata, plot_method, background, clustering,
            background_legend, cmap, cluster_cmap, summary, ndsize, scale,
            grid_density, grid_knn, grid_scale, grid_thresh, grid_width,
            stream_density, stream_linewidth, stream_cutoff_perc,
            arrow_color, filename, ax
        )

def create_3d_html_plot(X, V, signal_sum, adata, plot_method, background, clustering, 
                        cmap, cluster_cmap, summary, ndsize, scale,
                        grid_density, grid_knn, grid_scale, grid_thresh,
                        stream_density, stream_linewidth, stream_cutoff_perc,
                        arrow_color, filename, camera_eye, background_legend,
                        cutoff_perc=90):
    """创建3D HTML可视化 - 支持所有绘图模式"""
    
    fig = go.Figure()
    
    # 添加背景细胞点
    add_3d_background(fig, X, signal_sum, adata, background, clustering, 
                      cmap, cluster_cmap, ndsize, background_legend)
    
    # 根据绘图方法添加向量可视化
    if plot_method == "cell":
        add_3d_cell_vectors(fig, X, V, scale, arrow_color)
    elif plot_method == "grid":
        add_3d_grid_vectors(fig, X, V, grid_density, grid_knn, grid_scale, 
                           grid_thresh, scale, arrow_color, cutoff_perc)
    elif plot_method == "stream":
        add_3d_streamlines(fig, X, V, stream_density, stream_linewidth, 
                          stream_cutoff_perc, arrow_color)
    
    # 配置3D场景
    configure_3d_scene(fig, camera_eye, filename)
    
    return fig

def add_3d_background(fig, X, signal_sum, adata, background, clustering, 
                      cmap, cluster_cmap, ndsize, background_legend):
    """添加3D背景细胞点"""
    
    if background == 'summary':
        # 按信号强度着色
        print(f"Signal sum range: {signal_sum.min():.3f} - {signal_sum.max():.3f}")
        fig.add_trace(go.Scatter3d(
            x=X[:, 0], y=X[:, 1], z=X[:, 2],
            mode='markers',
            marker=dict(
                size=ndsize * 5,
                color=signal_sum,
                colorscale=cmap,
                opacity=0.8,
                line=dict(width=0),
                colorbar=dict(
                    title="Signal Strength",
                    titleside="right",
                    thickness=15,
                    len=0.7
                ),
                cmin=signal_sum.min(),
                cmax=signal_sum.max()
            ),
            name='Cells (Signal Strength)',
            showlegend=background_legend,
            text=[f'Cell {i}<br>Signal: {signal_sum[i]:.3f}' for i in range(len(signal_sum))],
            hovertemplate='<b>%{text}</b><br>Position: (%{x:.2f}, %{y:.2f}, %{z:.2f})<extra></extra>'
        ))
    
    elif background == 'cluster' and clustering:
        # 按细胞群着色
        labels = np.array(adata.obs[clustering], str)
        unique_labels = np.unique(labels)
        
        # 获取颜色映射
        if cluster_cmap is None:
            colors = get_plotly_colors(cmap, len(unique_labels))
            cluster_cmap = {label: colors[i] for i, label in enumerate(unique_labels)}
        
        for label in unique_labels:
            idx = labels == label
            fig.add_trace(go.Scatter3d(
                x=X[idx, 0], y=X[idx, 1], z=X[idx, 2],
                mode='markers',
                marker=dict(
                    size=ndsize * 3,
                    color=cluster_cmap[label],
                    opacity=0.6,
                    line=dict(width=0)
                ),
                name=f'Cluster {label}',
                showlegend=background_legend,
                text=[f'Cell {i}<br>Cluster: {label}' for i in np.where(idx)[0]],
                hovertemplate='<b>%{text}</b><br>(%{x:.2f}, %{y:.2f}, %{z:.2f})<extra></extra>'
            ))

def add_3d_cell_vectors(fig, X, V, scale, arrow_color):
    """添加3D细胞级向量（cell模式）"""
    
    # 过滤有效向量
    vector_magnitude = np.linalg.norm(V, axis=1)
    valid_mask = vector_magnitude > np.percentile(vector_magnitude[vector_magnitude > 0], 10)
    
    if np.any(valid_mask):
        X_valid = X[valid_mask]
        V_valid = V[valid_mask]
        
        fig.add_trace(go.Cone(
            x=X_valid[:, 0], y=X_valid[:, 1], z=X_valid[:, 2],
            u=V_valid[:, 0] * scale, v=V_valid[:, 1] * scale, w=V_valid[:, 2] * scale,
            sizemode="absolute",
            sizeref=0.5,
            colorscale=[[0, arrow_color], [1, arrow_color]],
            showscale=False,
            opacity=0.8,
            name='Cell Communication Vectors',
            hovertemplate='Cell Vector<br>Position: (%{x:.2f}, %{y:.2f}, %{z:.2f})<br>Direction: (%{u:.2f}, %{v:.2f}, %{w:.2f})<extra></extra>'
        ))
        print(f"Added {len(X_valid)} cell vectors")

def add_3d_grid_vectors(fig, X, V, grid_density, grid_knn, grid_scale, 
                       grid_thresh, scale, arrow_color, cutoff_perc=90):
    """添加3D网格向量（grid模式）"""
    
    # 创建3D网格并插值
    X_grid, V_grid = create_3d_grid_interpolation(
        X, V, grid_density, grid_knn, grid_scale, grid_thresh
    )
    
    original_vector_count = len(X_grid)

    # 添加通信向量
    if original_vector_count > 0:
        # 新增：过滤掉小的向量来隐藏小箭头
        vector_magnitudes = np.linalg.norm(V_grid, axis=1)
        
        # 仅在有正值向量时进行过滤
        positive_magnitudes = vector_magnitudes[vector_magnitudes > 0]
        if len(positive_magnitudes) > 0:
            # 将阈值设置为20%，意味着最小的20%的箭头将被隐藏
            threshold = np.percentile(positive_magnitudes, cutoff_perc)
            mask = vector_magnitudes > threshold
            X_grid = X_grid[mask]
            V_grid = V_grid[mask]

        if len(X_grid) == 0:
            print("All grid vectors were filtered out due to small magnitude.")
            return
            
        fig.add_trace(go.Cone(
            x=X_grid[:, 0], y=X_grid[:, 1], z=X_grid[:, 2],
            u=V_grid[:, 0] * scale, v=V_grid[:, 1] * scale, w=V_grid[:, 2] * scale,
            sizemode="absolute",
            sizeref=1,
            colorscale=[[0, arrow_color], [1, arrow_color]],
            showscale=False,
            opacity=0.8,
            name='Grid Communication Vectors',
            hovertemplate='Grid Vector<br>Position: (%{x:.2f}, %{y:.2f}, %{z:.2f})<br>Direction: (%{u:.2f}, %{v:.2f}, %{w:.2f})<extra></extra>'
        ))
        print(f"Added {len(X_grid)} grid vectors (filtered from {original_vector_count})")

def add_3d_streamlines(fig, X, V, stream_density, stream_linewidth, 
                      stream_cutoff_perc, arrow_color):
    """添加3D流线（stream模式）- 使用Plotly Streamtube"""
    
    print("Generating 3D streamtubes...")
    
    # 方法1: 使用Plotly原生Streamtube (推荐)
    try:
        add_plotly_streamtubes(fig, X, V, stream_density, stream_linewidth, arrow_color)
        print("✅ Using Plotly Streamtubes")
        return
    except Exception as e:
        print(f"⚠️ Streamtube failed ({e}), falling back to manual streamlines")
    
    # 方法2: 备用 - 手动绘制流线
    add_manual_streamlines(fig, X, V, stream_density, stream_linewidth, 
                          stream_cutoff_perc, arrow_color)

def add_plotly_streamtubes(fig, X, V, stream_density, stream_linewidth, arrow_color):
    """使用Plotly原生Streamtube功能"""
    
    # 创建规则网格用于Streamtube
    X_grid, V_grid = create_regular_3d_grid(X, V, stream_density)
    
    if X_grid is None:
        raise ValueError("Cannot create regular grid for streamtubes")
    
    # 获取网格尺寸
    x_unique = np.unique(X_grid[:, 0])
    y_unique = np.unique(X_grid[:, 1])  
    z_unique = np.unique(X_grid[:, 2])
    
    nx, ny, nz = len(x_unique), len(y_unique), len(z_unique)
    
    # 重塑向量场为网格形状
    U = V_grid[:, 0].reshape((nx, ny, nz))
    V_comp = V_grid[:, 1].reshape((nx, ny, nz))
    W = V_grid[:, 2].reshape((nx, ny, nz))
    
    # 创建起始点
    starts_x, starts_y, starts_z = create_streamtube_starts(x_unique, y_unique, z_unique, stream_density)
    
    # 添加Streamtube
    fig.add_trace(go.Streamtube(
        x=x_unique,
        y=y_unique, 
        z=z_unique,
        u=U,
        v=V_comp,
        w=W,
        starts=dict(
            x=starts_x,
            y=starts_y,
            z=starts_z
        ),
        sizeref=stream_linewidth * 0.1,
        colorscale=[[0, arrow_color], [1, arrow_color]],
        showscale=False,
        maxdisplayed=int(100 * stream_density),  # 控制显示的流管数量
        name='Communication Streamtubes',
        hovertemplate='Streamtube<br>Position: (%{x:.2f}, %{y:.2f}, %{z:.2f})<extra></extra>'
    ))

def create_regular_3d_grid(X, V, density=1.0):
    """为Streamtube创建规则网格"""
    
    # 计算空间边界
    x_min, x_max = X[:, 0].min(), X[:, 0].max()
    y_min, y_max = X[:, 1].min(), X[:, 1].max()
    z_min, z_max = X[:, 2].min(), X[:, 2].max()
    
    # 扩展边界
    padding = 0.1
    x_range, y_range, z_range = x_max - x_min, y_max - y_min, z_max - z_min
    x_min -= padding * x_range
    x_max += padding * x_range
    y_min -= padding * y_range  
    y_max += padding * y_range
    z_min -= padding * z_range
    z_max += padding * z_range
    
    # 创建规则网格
    n_grid = max(8, int(12 * density))
    
    x_grid = np.linspace(x_min, x_max, n_grid)
    y_grid = np.linspace(y_min, y_max, n_grid)
    z_grid = np.linspace(z_min, z_max, max(6, n_grid//2))
    
    mesh_x, mesh_y, mesh_z = np.meshgrid(x_grid, y_grid, z_grid, indexing='ij')
    grid_points = np.column_stack([mesh_x.flatten(), mesh_y.flatten(), mesh_z.flatten()])
    
    # 使用K近邻插值向量场到规则网格
    nbrs = NearestNeighbors(n_neighbors=min(8, len(X)), algorithm='ball_tree')
    nbrs.fit(X)
    distances, indices = nbrs.kneighbors(grid_points)
    
    # 高斯权重插值
    sigma = min(x_range, y_range, z_range) / n_grid
    weights = np.exp(-0.5 * (distances / sigma) ** 2)
    weight_sums = weights.sum(axis=1)
    
    V_grid = np.zeros([len(grid_points), 3])
    for dim in range(3):
        if dim < V.shape[1]:
            V_values = V[indices, dim]
            V_grid[:, dim] = (V_values * weights).sum(axis=1) / np.maximum(weight_sums, 1e-10)
    
    return grid_points, V_grid

def create_streamtube_starts(x_unique, y_unique, z_unique, density):
    """创建Streamtube起始点"""
    
    # 在边界面创建起始点
    n_starts_per_face = max(3, int(8 * density))
    
    starts_x, starts_y, starts_z = [], [], []
    
    # X边界面
    for x_val in [x_unique[0], x_unique[-1]]:
        y_starts = np.linspace(y_unique[0], y_unique[-1], n_starts_per_face)
        z_starts = np.linspace(z_unique[0], z_unique[-1], n_starts_per_face)
        yy, zz = np.meshgrid(y_starts, z_starts)
        starts_x.extend([x_val] * len(yy.flatten()))
        starts_y.extend(yy.flatten())
        starts_z.extend(zz.flatten())
    
    # Y边界面
    for y_val in [y_unique[0], y_unique[-1]]:
        x_starts = np.linspace(x_unique[0], x_unique[-1], n_starts_per_face)
        z_starts = np.linspace(z_unique[0], z_unique[-1], n_starts_per_face)
        xx, zz = np.meshgrid(x_starts, z_starts)
        starts_x.extend(xx.flatten())
        starts_y.extend([y_val] * len(xx.flatten()))
        starts_z.extend(zz.flatten())
    
    return starts_x, starts_y, starts_z

def add_manual_streamlines(fig, X, V, stream_density, stream_linewidth, 
                          stream_cutoff_perc, arrow_color):
    """备用方案：手动绘制流线"""
    
    # 创建流线起始点
    streamline_starts = create_3d_streamline_starts(X, V, stream_density, stream_cutoff_perc)
    
    if len(streamline_starts) == 0:
        print("No valid streamline starting points found")
        return
    
    # 为每个起始点生成流线
    all_streamlines = []
    
    for start_point in streamline_starts:
        streamline = integrate_3d_streamline(start_point, X, V)
        if len(streamline) > 2:  # 至少需要3个点
            all_streamlines.append(streamline)
    
    print(f"Generated {len(all_streamlines)} manual streamlines")
    
    # 绘制流线
    for i, streamline in enumerate(all_streamlines):
        fig.add_trace(go.Scatter3d(
            x=streamline[:, 0], 
            y=streamline[:, 1], 
            z=streamline[:, 2],
            mode='lines',
            line=dict(
                color=arrow_color,
                width=stream_linewidth * 2,
            ),
            opacity=0.7,
            name=f'Streamline {i+1}' if i < 5 else '',
            showlegend=i < 5,
            hovertemplate='Streamline<br>(%{x:.2f}, %{y:.2f}, %{z:.2f})<extra></extra>'
        ))
    
    # 添加流线起始点
    fig.add_trace(go.Scatter3d(
        x=streamline_starts[:, 0], 
        y=streamline_starts[:, 1], 
        z=streamline_starts[:, 2],
        mode='markers',
        marker=dict(
            size=3,
            color='red',
            opacity=0.8
        ),
        name='Streamline Origins',
        showlegend=True,
        hovertemplate='Streamline Start<br>(%{x:.2f}, %{y:.2f}, %{z:.2f})<extra></extra>'
    ))

def create_3d_streamline_starts(X, V, stream_density, stream_cutoff_perc):
    """创建3D流线起始点"""
    
    # 计算向量强度
    vector_magnitude = np.linalg.norm(V, axis=1)
    
    # 过滤掉弱向量
    magnitude_threshold = np.percentile(vector_magnitude, stream_cutoff_perc)
    strong_vector_mask = vector_magnitude > magnitude_threshold
    
    if not np.any(strong_vector_mask):
        return np.array([]).reshape(0, 3)
    
    # 在强向量区域创建起始点
    X_strong = X[strong_vector_mask]
    
    # 根据密度参数采样起始点
    n_starts = max(5, int(len(X_strong) * stream_density * 0.1))
    n_starts = min(n_starts, len(X_strong))
    
    # 使用k-means样采样或均匀采样
    if n_starts < len(X_strong):
        indices = np.random.choice(len(X_strong), n_starts, replace=False)
        streamline_starts = X_strong[indices]
    else:
        streamline_starts = X_strong
    
    print(f"Created {len(streamline_starts)} streamline starting points")
    return streamline_starts

def integrate_3d_streamline(start_point, X, V, max_length=50, step_size=0.1):
    """积分生成3D流线"""
    
    # 使用K近邻插值向量场
    def vector_field_interpolator(pos):
        pos = pos.reshape(1, -1)
        
        # 找到最近的邻居
        nbrs = NearestNeighbors(n_neighbors=min(5, len(X)), algorithm='ball_tree')
        nbrs.fit(X)
        distances, indices = nbrs.kneighbors(pos)
        
        # 使用距离权重插值
        weights = 1.0 / (distances[0] + 1e-10)
        weights = weights / weights.sum()
        
        # 插值向量
        interpolated_v = np.sum(V[indices[0]] * weights[:, None], axis=0)
        return interpolated_v
    
    # 积分流线
    streamline = [start_point.copy()]
    current_pos = start_point.copy()
    
    for step in range(max_length):
        # 获取当前位置的向量
        try:
            current_v = vector_field_interpolator(current_pos)
            
            # 检查向量是否有效
            if np.linalg.norm(current_v) < 1e-6:
                break
            
            # 归一化并应用步长
            current_v = current_v / np.linalg.norm(current_v) * step_size
            
            # 更新位置
            next_pos = current_pos + current_v
            streamline.append(next_pos.copy())
            current_pos = next_pos
            
            # 检查是否超出边界
            X_min, X_max = X.min(axis=0), X.max(axis=0)
            if np.any(current_pos < X_min - 0.5*(X_max-X_min)) or np.any(current_pos > X_max + 0.5*(X_max-X_min)):
                break
                
        except Exception as e:
            print(f"Streamline integration error: {e}")
            break
    
    return np.array(streamline)

def create_3d_grid_interpolation(X, V, grid_density, grid_knn, grid_scale, grid_thresh):
    """创建3D网格插值"""
    
    # 计算空间范围
    mins = np.min(X, axis=0)
    maxs = np.max(X, axis=0)
    ranges = maxs - mins
    
    # 扩展边界
    padding = 0.1
    mins -= padding * ranges
    maxs += padding * ranges
    
    # 创建3D网格
    n_grid = max(8, int(15 * grid_density))  # 控制网格密度
    
    if X.shape[1] == 3:
        # 真3D网格
        x_grid = np.linspace(mins[0], maxs[0], n_grid)
        y_grid = np.linspace(mins[1], maxs[1], n_grid)
        z_grid = np.linspace(mins[2], maxs[2], n_grid)
        mesh_x, mesh_y, mesh_z = np.meshgrid(x_grid, y_grid, z_grid, indexing='ij')
        grid_points = np.column_stack([mesh_x.flatten(), mesh_y.flatten(), mesh_z.flatten()])
    else:
        # 2D数据扩展到3D
        x_grid = np.linspace(mins[0], maxs[0], n_grid)
        y_grid = np.linspace(mins[1], maxs[1], n_grid)
        z_grid = np.linspace(0, ranges.mean() * 0.5, max(4, n_grid//3))
        mesh_x, mesh_y, mesh_z = np.meshgrid(x_grid, y_grid, z_grid, indexing='ij')
        grid_points = np.column_stack([mesh_x.flatten(), mesh_y.flatten(), mesh_z.flatten()])
    
    # K近邻插值
    if grid_knn is None:
        grid_knn = max(8, min(50, X.shape[0] // 10))
    
    # 确保有足够的邻居
    grid_knn = min(grid_knn, X.shape[0])
    
    nbrs = NearestNeighbors(n_neighbors=grid_knn, algorithm='ball_tree')
    nbrs.fit(X)
    distances, indices = nbrs.kneighbors(grid_points)
    
    # 高斯权重
    grid_size = np.mean(ranges) / n_grid
    sigma = grid_size * grid_scale
    weights = np.exp(-0.5 * (distances / sigma) ** 2)
    weight_sums = weights.sum(axis=1)
    
    # 插值向量
    V_grid = np.zeros([len(grid_points), V.shape[1]])
    for dim in range(V.shape[1]):
        V_values = V[indices, dim]  # shape: (n_grid_points, grid_knn)
        V_grid[:, dim] = (V_values * weights).sum(axis=1) / np.maximum(weight_sums, 1e-10)
    
    # 应用阈值
    threshold = np.percentile(weight_sums, 50) * grid_thresh  # 降低阈值
    valid_mask = weight_sums > threshold
    
    # 过滤掉零向量
    vector_magnitude = np.linalg.norm(V_grid, axis=1)
    magnitude_threshold = np.percentile(vector_magnitude[vector_magnitude > 0], 10) if np.any(vector_magnitude > 0) else 0
    valid_mask &= vector_magnitude > magnitude_threshold
    
    return grid_points[valid_mask], V_grid[valid_mask]

def configure_3d_scene(fig, camera_eye, filename):
    """配置3D场景"""
    
    scene_config = dict(
        xaxis=dict(title='X', showgrid=True),
        yaxis=dict(title='Y', showgrid=True),
        zaxis=dict(title='Z', showgrid=True),
        aspectmode='cube'
    )
    
    if camera_eye:
        scene_config['camera'] = dict(eye=camera_eye)
    else:
        scene_config['camera'] = dict(eye=dict(x=1.5, y=1.5, z=1.5))
    
    fig.update_layout(
        title='3D Cell Communication',
        scene=scene_config,
        width=1000,
        height=800,
        margin=dict(l=0, r=0, t=50, b=0)
    )
    
    # 保存HTML文件
    if filename:
        html_filename = filename if filename.endswith('.html') else filename + '.html'
        fig.write_html(html_filename, include_plotlyjs=True)
        print(f"3D visualization saved as '{html_filename}'")

def get_plotly_colors(colormap_name, n_colors):
    """获取Plotly颜色映射 - 兼容版本"""
    try:
        color_maps = {
            'Plotly': pq.Plotly,
            'Light24': pq.Light24, 
            'Dark24': pq.Dark24,
            'Alphabet': pq.Alphabet
        }
        colors = color_maps.get(colormap_name, pq.Plotly)
    except AttributeError:
        # 备用颜色方案
        backup_colors = {
            'Plotly': ['#636EFA', '#EF553B', '#00CC96', '#AB63FA', '#FFA15A', '#19D3F3', '#FF6692', '#B6E880', '#FF97FF', '#FECB52'],
            'Light24': ['#FD3216', '#00FE35', '#6A76FC', '#FED4C4', '#FE00CE', '#0DF9FF', '#F6F926', '#FF9616', '#479B55', '#EEA6FB'],
            'Dark24': ['#2E91E5', '#E15F99', '#1CA71C', '#FB0D0D', '#DA16FF', '#222A2A', '#B68100', '#750D86', '#EB663B', '#511CFB'],
            'Alphabet': ['#AA0DFE', '#3283FE', '#85660D', '#782AB6', '#565656', '#1C8356', '#16FF32', '#F7E1A0', '#E2E2E2', '#1CBE4F']
        }
        colors = backup_colors.get(colormap_name, backup_colors['Plotly'])
    
    return [colors[i % len(colors)] for i in range(n_colors)]

def plot_original_2d(X, V, signal_sum, adata, plot_method, background, clustering,
                     background_legend, cmap, cluster_cmap, summary, ndsize, scale,
                     grid_density, grid_knn, grid_scale, grid_thresh, grid_width,
                     stream_density, stream_linewidth, stream_cutoff_perc,
                     arrow_color, filename, ax):
    """原始2D绘图逻辑的简化版本"""
    
    import matplotlib.pyplot as plt
    
    if ax is None:
        fig, ax = plt.subplots(figsize=(10, 8))
    
    # 背景点
    if background == 'summary':
        scatter = ax.scatter(X[:, 0], X[:, 1], s=ndsize*20, c=signal_sum, 
                           cmap=cmap, linewidth=0, alpha=0.7)
        if background_legend:
            plt.colorbar(scatter, ax=ax, label='Signal Strength')
    
    elif background == 'cluster' and clustering:
        labels = adata.obs[clustering].astype('category')
        unique_labels = labels.cat.categories
        
        for i, label in enumerate(unique_labels):
            mask = labels == label
            color = cluster_cmap[label] if cluster_cmap and label in cluster_cmap else f'C{i}'
            ax.scatter(X[mask, 0], X[mask, 1], s=ndsize*20, c=color, 
                      label=label, linewidth=0, alpha=0.7)
        
        if background_legend:
            ax.legend(bbox_to_anchor=(1.05, 1), loc='upper left')
    
    # 添加向量
    if plot_method == "cell":
        # 过滤零向量
        V_plot = V.copy()
        valid_mask = np.linalg.norm(V_plot, axis=1) > 1e-6
        if np.any(valid_mask):
            ax.quiver(X[valid_mask, 0], X[valid_mask, 1], 
                     V_plot[valid_mask, 0], V_plot[valid_mask, 1],
                     scale=scale, scale_units='xy', angles='xy',
                     color=arrow_color, alpha=0.8, width=0.003)
    
    elif plot_method == "grid":
        # 2D网格插值 (简化版)
        X_grid, V_grid = create_2d_grid_interpolation(X[:, :2], V[:, :2], grid_density, grid_knn, grid_scale, grid_thresh)
        if len(X_grid) > 0:
            ax.quiver(X_grid[:, 0], X_grid[:, 1], V_grid[:, 0], V_grid[:, 1],
                     scale=scale, scale_units='xy', angles='xy',
                     color=arrow_color, alpha=0.8, width=grid_width)
    
    ax.set_aspect('equal')
    ax.axis('off')
    
    if filename:
        plt.savefig(filename, dpi=300, bbox_inches='tight', transparent=True)
    
    return ax

def create_2d_grid_interpolation(X, V, grid_density, grid_knn, grid_scale, grid_thresh):
    """2D网格插值"""
    x_min, x_max = X[:, 0].min(), X[:, 0].max()
    y_min, y_max = X[:, 1].min(), X[:, 1].max()
    
    x_range = x_max - x_min
    y_range = y_max - y_min
    
    n_grid = int(30 * grid_density)
    x_grid = np.linspace(x_min - 0.1*x_range, x_max + 0.1*x_range, n_grid)
    y_grid = np.linspace(y_min - 0.1*y_range, y_max + 0.1*y_range, n_grid)
    
    mesh_x, mesh_y = np.meshgrid(x_grid, y_grid)
    grid_points = np.column_stack([mesh_x.flatten(), mesh_y.flatten()])
    
    if grid_knn is None:
        grid_knn = max(5, X.shape[0] // 20)
    grid_knn = min(grid_knn, X.shape[0])
    
    nbrs = NearestNeighbors(n_neighbors=grid_knn)
    nbrs.fit(X)
    distances, indices = nbrs.kneighbors(grid_points)
    
    grid_size = min(x_range, y_range) / n_grid
    weights = np.exp(-0.5 * (distances / (grid_size * grid_scale)) ** 2)
    weight_sums = weights.sum(axis=1)
    
    V_grid = np.zeros([len(grid_points), V.shape[1]])
    for dim in range(V.shape[1]):
        V_grid[:, dim] = (V[indices, dim] * weights).sum(axis=1) / np.maximum(weight_sums, 1e-10)
    
    threshold = np.percentile(weight_sums, 70) * grid_thresh
    valid_mask = weight_sums > threshold
    valid_mask &= np.linalg.norm(V_grid, axis=1) > np.percentile(np.linalg.norm(V_grid, axis=1), 20)
    
    return grid_points[valid_mask], V_grid[valid_mask]

# 使用示例 - 现在支持所有3D模式
def example_usage(adata):
    """使用示例 - 演示所有3D模式"""
    
    print("=== 生成所有3D模式的可视化 ===")
    
    # 1. Cell模式 - 3D
    print("\n1. 生成3D Cell模式...")
    fig_cell = plot_cell_communication(
        adata,
        database_name='cellchat',
        plot_method='cell',  # cell模式
        background='summary',
        pos_idx=np.array([0, 1, 2]),
        html_3d=True,
        background_legend=True,
        filename='cell_comm_3d_cell.html',
        camera_eye={'x': 2, 'y': 2, 'z': 1.5},
        scale=2.0  # 放大箭头便于观察
    )
    
    # 2. Grid模式 - 3D  
    print("\n2. 生成3D Grid模式...")
    fig_grid = plot_cell_communication(
        adata,
        database_name='cellchat',
        plot_method='grid',  # grid模式
        background='summary',
        pos_idx=np.array([0, 1, 2]),
        html_3d=True,
        background_legend=True,
        filename='cell_comm_3d_grid.html',
        camera_eye={'x': 2, 'y': 2, 'z': 1.5},
        grid_density=1.2
    )
    
    # 3. Stream模式 - 3D (使用Plotly Streamtubes!)
    print("\n3. 生成3D Stream模式 (Streamtubes)...")
    fig_stream = plot_cell_communication(
        adata,
        database_name='cellchat',
        plot_method='stream',  # stream模式 - 现在使用Streamtubes!
        background='cluster',
        clustering='leiden',
        pos_idx=np.array([0, 1, 2]),
        html_3d=True,
        background_legend=True,
        filename='cell_comm_3d_streamtubes.html',
        camera_eye={'x': 2, 'y': 2, 'z': 1.5},
        stream_density=1.0,    # 更好的密度控制
        stream_linewidth=3,    # Streamtube粗细
        stream_cutoff_perc=20
    )
    
    print("\n=== 完成！生成的文件：===")
    print("- cell_comm_3d_cell.html        (3D Cell模式 - 单细胞向量)")
    print("- cell_comm_3d_grid.html        (3D Grid模式 - 网格插值)")  
    print("- cell_comm_3d_streamtubes.html (3D Stream模式 - Plotly Streamtubes!)")
    print("\n🎯 Stream模式特别说明：")
    print("- 现在使用Plotly原生Streamtube功能")
    print("- 具有体积感的3D流管可视化")
    print("- 自动流线积分和渲染")
    print("- 如果Streamtube失败，会自动降级到手动流线")
    print("\n在浏览器中打开HTML文件，可以:")
    print("- 拖拽旋转3D场景")
    print("- 鼠标悬停查看详细信息")
    print("- 滚轮缩放")
    print("- 使用图例控制显示/隐藏元素")
    
    return fig_cell, fig_grid, fig_stream

# 快速测试函数
def test_all_3d_modes(adata, database_name='cellchat'):
    """快速测试所有3D模式"""
    
    print("Testing all 3D visualization modes...")
    
    modes = ['cell', 'grid', 'stream']
    for mode in modes:
        print(f"\nTesting {mode} mode...")
        try:
            fig = plot_cell_communication(
                adata,
                database_name=database_name,
                plot_method=mode,
                background='summary',
                pos_idx=np.array([0, 1, 2]),
                html_3d=True,
                filename=f'test_3d_{mode}_streamtubes.html' if mode == 'stream' else f'test_3d_{mode}.html',
                scale=1.5 if mode == 'cell' else 1.0,
                stream_density=0.5 if mode == 'stream' else 1.0
            )
            print(f"✅ {mode} mode successful")
        except Exception as e:
            print(f"❌ {mode} mode failed: {e}")
    
    print("\nTest completed!")

def test_signal_visualization(adata, database_name='cellchat', summary='sender'):
    """快速测试signal_sum可视化是否正常"""
    
    # 检查数据
    if summary == 'sender':
        signal_key = 's-total-total'
        signal_data = adata.obsm[f'commot-{database_name}-sum-sender']
    else:
        signal_key = 'r-total-total' 
        signal_data = adata.obsm[f'commot-{database_name}-sum-receiver']
    
    if signal_key in signal_data.columns:
        signal_values = signal_data[signal_key].values
        print(f"✅ Signal data found: {signal_key}")
        print(f"   Range: {signal_values.min():.3f} - {signal_values.max():.3f}")
        print(f"   Mean: {signal_values.mean():.3f}")
        print(f"   Non-zero values: {np.sum(signal_values > 0)}/{len(signal_values)}")
        return signal_values
    else:
        print(f"❌ Signal data NOT found: {signal_key}")
        print(f"   Available keys: {list(signal_data.columns)}")
        return None


# -*- coding: utf-8 -*-

import logging
from ctypes import c_uint32
from math import ceil
from multiprocessing import Array, Process, cpu_count
from multiprocessing.sharedctypes import RawArray
from operator import attrgetter, mul
from typing import Sequence, Type

import numpy as np
import pandas as pd
from boltons.iterutils import chunked
from tqdm import tqdm
from scipy.sparse import csr_matrix

LOGGER = logging.getLogger(__name__)
# To reduce the memory footprint of a ranking matrix we use unsigned 32bit integers which provides a range from 0
# through 4,294,967,295. This should be sufficient even for region-based approaches.
DTYPE = "uint32"
DTYPE_C = c_uint32

ctxcore_install=False

def check_ctxcore():
    r"""Check if ctxcore package is installed for AUCell analysis.
    
    Raises:
        ImportError: If ctxcore is not installed
    """
    global ctxcore_install
    try:
        import ctxcore
        ctxcore_install=True
        #print('ctxcore have been install version:',ctxcore.__version__)
    except ImportError:
        raise ImportError(
            'Please install the ctxcore: `pip install ctxcore`.'
        )

def global_imports(modulename,shortname = None, asfunction = False):
    if shortname is None: 
        shortname = modulename
    if asfunction is False:
        globals()[shortname] = __import__(modulename)
    else:        
        globals()[shortname] = __import__(modulename)

def create_rankings(ex_mtx: pd.DataFrame, seed=None) -> pd.DataFrame:
    r"""Create genome-wide gene rankings from single-cell expression data.

    This function ranks genes for each cell based on expression levels,
    creating rankings suitable for AUCell gene set enrichment analysis.

    Arguments:
        ex_mtx: Expression matrix with cells as rows and genes as columns (n_cells x n_genes)
        seed (int): Random seed for reproducible ranking (default: None)
    
    Returns:
        pd.DataFrame: Gene rankings matrix with same dimensions as input (n_cells x n_genes)

    """
    # Do a shuffle would be nice for exactly similar behaviour as R implementation.
    # 1. Ranks are assigned in the range of 1 to n, therefore we need to subtract 1.
    # 2. In case of a tie the 'first' method is used, i.e. we keep the order in the original array. The remove any
    #    bias we shuffle the dataframe before ranking it. This introduces a performance penalty!
    # 3. Genes are ranked according to gene expression in descending order, i.e. from highly expressed (0) to low expression (n).
    # 3. NAs should be given the highest rank numbers. Documentation is bad, so tested implementation via code snippet:
    #
    #    import pandas as pd
    #    import numpy as np
    #    df = pd.DataFrame(data=[4, 1, 3, np.nan, 2, 3], columns=['values'])
    #    # Run below statement multiple times to see effect of shuffling in case of a tie.
    #    df.sample(frac=1.0, replace=False).rank(ascending=False, method='first', na_option='bottom').sort_index() - 1
    #
    return (
        ex_mtx.sample(frac=1.0, replace=False, axis=1, random_state=seed)
        .rank(axis=1, ascending=False, method="first", na_option="bottom")
        .astype(DTYPE)
        - 1
    )

def _rank_sparse_row(row_sparse, n_cols):
    row = row_sparse.toarray().ravel()
    sort_idx = np.lexsort((np.arange(n_cols), -row))
    rank_vals = np.empty(n_cols, dtype=np.int32)
    rank_vals[sort_idx] = np.arange(n_cols)
    return rank_vals

def fast_rank(X_sparse, seed=42):

    rng = np.random.default_rng(seed)
    n_rows, n_cols = X_sparse.shape

    shuffle_order = rng.permutation(n_cols)
    X_shuffled = X_sparse[:, shuffle_order]

    # results = Parallel(n_jobs=n_jobs, backend=backend)(
    #     delayed(_rank_sparse_row)(X_shuffled.getrow(i), n_cols) for i in range(n_rows)
    # )
    results = []
    for i in tqdm(range(n_rows)):
        row_result = _rank_sparse_row(X_shuffled.getrow(i), n_cols)
        results.append(row_result)
    ranks = np.stack(results)

    # Undo column shuffle
    unshuffled_ranks = np.empty_like(ranks)
    unshuffled_ranks[:, shuffle_order] = ranks

    return unshuffled_ranks



def derive_auc_threshold(ex_mtx: csr_matrix) -> pd.DataFrame:
    """
    Derive AUC thresholds for an expression matrix.

    It is important to check that most cells have a substantial fraction of expressed/detected genes in the calculation of
    the AUC.
    
    Arguments:
        ex_mtx: The expression profile matrix. The rows should correspond to different cells, the columns to different genes (n_cells x n_genes).
    
    Returns:
        A dataframe with AUC threshold for different quantiles over the number cells: a fraction of 0.01 designates that when using this value as the AUC threshold for 99% of the cells all ranked genes used for AUC calculation will have had a detected expression in the single-cell experiment.

    """
    return (
        pd.Series(np.count_nonzero(ex_mtx.toarray(), axis=1)).quantile(
            [0.01, 0.05, 0.10, 0.50, 1]
        )
        / ex_mtx.shape[1]
    )



def _enrichment(
    shared_ro_memory_array, modules, genes, cells, auc_threshold, auc_mtx, offset
):
    check_ctxcore()
    global ctxcore_install
    if ctxcore_install==True:
        global enrichment4cells
        from ctxcore.recovery import enrichment4cells

    # The rankings dataframe is properly reconstructed (checked this).
    df_rnk = pd.DataFrame(
        data=np.frombuffer(shared_ro_memory_array, dtype=DTYPE).reshape(
            len(cells), len(genes)
        ),
        columns=genes,
        index=cells,
    )
    # To avoid additional memory burden de resulting AUCs are immediately stored in the output sync. array.
    result_mtx = np.frombuffer(auc_mtx.get_obj(), dtype="d")
    inc = len(cells)
    for idx, module in enumerate(modules):
        result_mtx[
            offset + (idx * inc) : offset + ((idx + 1) * inc)
        ] = enrichment4cells(df_rnk, module, auc_threshold).values.ravel(order="C")


def aucell4r(
    df_rnk: pd.DataFrame,
    signatures,
    auc_threshold: float = 0.05,
    noweights: bool = False,
    normalize: bool = False,
    num_workers: int = cpu_count(),
) -> pd.DataFrame:
    """
    Calculate enrichment of gene signatures for single cells.

    Arguments:
        df_rnk: The rank matrix (n_cells x n_genes).
        signatures: The gene signatures or regulons.
        auc_threshold: The fraction of the ranked genome to take into account for the calculation of the Area Under the recovery Curve.
        noweights: Should the weights of the genes part of a signature be used in calculation of enrichment?
        normalize: Normalize the AUC values to a maximum of 1.0 per regulon.
        num_workers: The number of cores to use.
    
    Returns:
        A dataframe with the AUCs (n_cells x n_modules).

    """
    check_ctxcore()
    global ctxcore_install
    if ctxcore_install==True:
        global enrichment4cells
        from ctxcore.recovery import enrichment4cells


    if num_workers == 1:
        # Show progress bar for pathway processing
        print(f"Computing AUC scores for {len(signatures)} pathways using single worker...")
        aucs = pd.concat(
            [
                enrichment4cells(
                    df_rnk,
                    module.noweights() if noweights else module,
                    auc_threshold=auc_threshold,
                )
                for module in tqdm(signatures, desc="Processing pathways")
            ]
        ).unstack("Regulon")
        aucs.columns = aucs.columns.droplevel(0)
    else:
        # Multi-worker processing with progress info
        print(f"Computing AUC scores for {len(signatures)} pathways using {num_workers} workers...")
        # Decompose the rankings dataframe: the index and columns are shared with the child processes via pickling.
        genes = df_rnk.columns.values
        cells = df_rnk.index.values
        # The actual rankings are shared directly. This is possible because during a fork from a parent process the child
        # process inherits the memory of the parent process. A RawArray is used instead of a synchronize Array because
        # these rankings are read-only.
        shared_ro_memory_array = RawArray(DTYPE_C, mul(*df_rnk.shape))
        array = np.frombuffer(shared_ro_memory_array, dtype=DTYPE)
        # Copy the contents of df_rank into this shared memory block using row-major ordering.
        array[:] = df_rnk.values.ravel(order="C")

        # The resulting AUCs are returned via a synchronize array.
        auc_mtx = Array("d", len(cells) * len(signatures))  # Double precision floats.

        # Convert the modules to modules with uniform weights if necessary.
        if noweights:
            signatures = list(map(lambda m: m.noweights(), signatures))

        # Do the analysis in separate child processes.
        chunk_size = ceil(float(len(signatures)) / num_workers)
        print(f"Splitting {len(signatures)} pathways into {num_workers} chunks of ~{chunk_size} pathways each...")
        
        processes = [
            Process(
                target=_enrichment,
                args=(
                    shared_ro_memory_array,
                    chunk,
                    genes,
                    cells,
                    auc_threshold,
                    auc_mtx,
                    (chunk_size * len(cells)) * idx,
                ),
            )
            for idx, chunk in enumerate(chunked(signatures, chunk_size))
        ]
        
        print("Starting parallel pathway processing...")
        for p in processes:
            p.start()
        for p in processes:
            p.join()
        print("Parallel processing completed!")

        # Reconstitute the results array. Using C or row-major ordering.
        aucs = pd.DataFrame(
            data=np.ctypeslib.as_array(auc_mtx.get_obj()).reshape(
                len(signatures), len(cells)
            ),
            columns=pd.Index(data=cells, name="Cell"),
            index=pd.Index(
                data=list(map(attrgetter("name"), signatures)), name="Regulon"
            ),
        ).T
    
    result = aucs / aucs.max(axis=0) if normalize else aucs
    print(f"AUC calculation completed! Generated scores for {result.shape[1]} pathways across {result.shape[0]} cells.")
    return result


def aucell(
    exp_mtx,
    signatures,
    auc_threshold: float = 0.05,
    noweights: bool = False,
    normalize: bool = False,
    seed=None,
    num_workers: int = cpu_count(),
    index=None,
    columns=None,
) -> pd.DataFrame:
    r"""Calculate gene signature enrichment scores using AUCell algorithm.

    AUCell quantifies gene set enrichment by calculating the Area Under the Curve (AUC)
    of gene rankings for each cell, providing a robust measure of pathway activity.

    Arguments:
        exp_mtx: Expression matrix (n_cells x n_genes) - DataFrame or sparse matrix
        signatures: Gene signatures or regulons for enrichment analysis
        auc_threshold (float): Fraction of ranked genes to consider for AUC calculation (default: 0.05)
        noweights (bool): Whether to ignore gene weights in signatures (default: False)
        normalize (bool): Whether to normalize AUC values to maximum of 1.0 per signature (default: False)
        seed (int): Random seed for reproducible gene ranking (default: None)
        num_workers (int): Number of CPU cores for parallel processing (default: all cores)
        index: Custom row index for output DataFrame (default: None)
        columns: Custom column names for output DataFrame (default: None)
    
    Returns:
        pd.DataFrame: AUC enrichment scores with cells as rows and signatures as columns

    """
    # Handle different input types
    if isinstance(exp_mtx, pd.DataFrame):
        # DataFrame input - extract index, columns, and convert to sparse
        original_index = exp_mtx.index
        original_columns = exp_mtx.columns
        matrix_values = exp_mtx.values
        if hasattr(matrix_values, 'toarray'):
            matrix_sparse = matrix_values
        else:
            from scipy.sparse import csr_matrix
            matrix_sparse = csr_matrix(matrix_values)
    else:
        # Sparse matrix input - use provided or default index/columns
        matrix_sparse = exp_mtx
        original_index = index if index is not None else pd.RangeIndex(exp_mtx.shape[0])
        original_columns = columns if columns is not None else pd.RangeIndex(exp_mtx.shape[1])
    
    # Use fast_rank for efficient ranking
    if seed is None:
        seed = 42
    rank_matrix = fast_rank(matrix_sparse, seed=seed)
    
    # Convert rank matrix back to DataFrame format for aucell4r
    df_rnk = pd.DataFrame(
        data=rank_matrix.astype(DTYPE),
        columns=original_columns,
        index=original_index
    )
    
    return aucell4r(
        df_rnk,
        signatures,
        auc_threshold,
        noweights,
        normalize,
        num_workers,
    )



import pandas as pd
import numpy as np
import plotly
import os
import matplotlib as mpl
import matplotlib.pyplot as plt
import networkx as nx
from sklearn.neighbors import NearestNeighbors
from sklearn.preprocessing import normalize
from scipy.stats import norm

def get_cmap_qualitative(cmap_name, n_colors=None):
    """
    Get qualitative colormap with support for large numbers of colors.
    
    Parameters
    ----------
    cmap_name : str
        Name of the colormap. Options include:
        - Standard plotly: "Plotly", "Alphabet", "Light24", "Dark24", "Safe", "Vivid"
        - Extended omicverse: "omicverse_28", "omicverse_56", "omicverse_112"  
        - Traditional Chinese: "forbidden_city"
        - Themed: "vibrant", "earth", "pastel", "ditto"
    n_colors : int, optional
        Number of colors needed. If provided, will automatically select 
        appropriate palette size.
    
    Returns
    -------
    cmap : list
        List of color strings
    """
    # Handle omicverse extended palettes directly
    if cmap_name in ["omicverse_28", "omicverse_56", "omicverse_112", "forbidden_city", 
                     "vibrant", "earth", "pastel", "ditto"]:
        try:
            from ...pl._palette import (palette_28, palette_56, palette_112, 
                                       vibrant_palette, earth_palette, pastel_palette, 
                                       ditto_color, ForbiddenCity)
            
            if cmap_name == "omicverse_28":
                cmap = palette_28
            elif cmap_name == "omicverse_56":
                cmap = palette_56  
            elif cmap_name == "omicverse_112":
                cmap = palette_112
            elif cmap_name == "vibrant":
                cmap = vibrant_palette
            elif cmap_name == "earth":
                cmap = earth_palette
            elif cmap_name == "pastel":
                cmap = pastel_palette
            elif cmap_name == "ditto":
                cmap = ditto_color
            elif cmap_name == "forbidden_city":
                fc = ForbiddenCity()
                # Use a diverse selection of traditional Chinese colors
                if n_colors is not None:
                    # Select colors from different color families for diversity
                    color_families = [
                        fc.green[:15], fc.red[:15], fc.blue[:15], fc.yellow[:15], 
                        fc.purple[:15], fc.brown[:15], fc.pink[:15], fc.grey[:10]
                    ]
                    cmap = []
                    family_idx = 0
                    color_idx = 0
                    while len(cmap) < n_colors and family_idx < len(color_families):
                        if color_idx < len(color_families[family_idx]):
                            cmap.append(color_families[family_idx][color_idx])
                            family_idx = (family_idx + 1) % len(color_families)
                            if family_idx == 0:
                                color_idx += 1
                        else:
                            family_idx += 1
                else:
                    cmap = fc.green[:10] + fc.red[:10] + fc.blue[:10] + fc.yellow[:10]
        except ImportError:
            # Fallback to plotly if import fails
            cmap = plotly.colors.qualitative.Plotly
    else:
        # Standard plotly palettes
        if cmap_name == "Plotly":
            cmap = plotly.colors.qualitative.Plotly
        elif cmap_name == "Alphabet":
            cmap = plotly.colors.qualitative.Alphabet
        elif cmap_name == "Light24":
            cmap = plotly.colors.qualitative.Light24
        elif cmap_name == "Dark24":
            cmap = plotly.colors.qualitative.Dark24
        elif cmap_name == "Safe":
            cmap = plotly.colors.qualitative.Safe
        elif cmap_name == "Vivid":
            cmap = plotly.colors.qualitative.Vivid
        else:
            cmap = plotly.colors.qualitative.Plotly
    
    # If n_colors is specified and exceeds current palette size, 
    # use extended palettes from omicverse
    if n_colors is not None and n_colors > len(cmap) and cmap_name not in ["forbidden_city"]:
        try:
            from ...pl._palette import palette_28, palette_56, palette_112
            
            if n_colors <= 28:
                cmap = palette_28[:n_colors]
            elif n_colors <= 56:
                cmap = palette_56[:n_colors]
            elif n_colors <= 112:
                cmap = palette_112[:n_colors]
            else:
                # For very large numbers, use palette_112 and cycle through
                base_palette = palette_112
                cycles_needed = (n_colors + len(base_palette) - 1) // len(base_palette)
                extended_palette = []
                for cycle in range(cycles_needed):
                    for color in base_palette:
                        if len(extended_palette) >= n_colors:
                            break
                        extended_palette.append(color)
                    if len(extended_palette) >= n_colors:
                        break
                cmap = extended_palette[:n_colors]
        except ImportError:
            # Fallback to original behavior if import fails
            pass
    
    return cmap

def linear_clamp_value(x, lower_bound, upper_bound, out_min, out_max):
    if x <= lower_bound:
        y = out_min
    elif x >= upper_bound:
        y = out_max
    else:
        y = out_min + (x - lower_bound)/(upper_bound-lower_bound) * (out_max-out_min)
    return y


def plot_cluster_signaling_chord(S, p_values, label_name=None, colormap='Plotly', quantile_cutoff=None, p_value_cutoff=None, cutoff=0, separate=False, filename="chord_plot.pdf", diagonal_off=False):
    cmap = get_cmap_qualitative(colormap, n_colors=S.shape[0])
    data_all = np.empty([0,3], str)
    if not quantile_cutoff is None:
        cutoff = np.quantile(S.reshape(-1), quantile_cutoff)
    for i in range(S.shape[0]):
        for j in range(S.shape[1]):
            if diagonal_off and i==j: continue
            if S[i,j] >= cutoff or p_values[i,j] <= p_value_cutoff:
                if not label_name is None:
                    if separate:
                        data = np.array([['S'+label_name[i], 'T'+label_name[j], str(S[i,j])]], str)
                    else:
                        data = np.array([[label_name[i], label_name[j], str(S[i,j])]], str)
                else:
                    if separate:
                        data = np.array([['S'+str(i), 'T'+str(j), str(S[i,j])]], str)
                    else:
                        data = np.array([[str(i), str(j), str(S[i,j])]], str)
                data_all = np.concatenate((data_all, data), axis=0)
    df = pd.DataFrame(data=data_all, columns=['from','to','value'])
    df.to_csv("./plot_chord_data.csv", index=False)

    data_cmap = np.empty([0, 2], str)
    for i in range(S.shape[0]):
        if not label_name is None:
            if separate:
                data_tmp = np.array([['S'+label_name[i], cmap[i % len(cmap)]],['T'+label_name[i], cmap[i % len(cmap)]]], str)
            else:
                data_tmp = np.array([[label_name[i], cmap[i % len(cmap)]]], str)
        else:
            if separate:
                data_tmp = np.array([['S'+str(i), cmap[i % len(cmap)]],['T'+str(i), cmap[i % len(cmap)]]], str)
            else:
                data_tmp = np.array([[str(i), cmap[i % len(cmap)]]], str)
        data_cmap = np.concatenate((data_cmap, data_tmp), axis=0)
    df = pd.DataFrame(data=data_cmap)
    df.to_csv("./plot_chord_cmap.csv", index=False, header=False)

    os.system("Rscript plot_chord.R")
    os.system("mv Rplots.pdf "+filename)
    os.system("pdfcrop %s %s" % (filename, filename))
    filename_png = filename[:-4]
    filename_png = filename_png + ".png"
    os.system("convert -density 800 %s -quality 100 %s" % (filename, filename_png))

def plot_cluster_signaling_network(S,
    labels = None,
    node_size = 0.2,
    node_colormap = "Plotly",
    node_cluster_colormap = None,
    node_pos = None,
    edge_width_lb_quantile = 0.05,
    edge_width_ub_quantile = 0.95,
    edge_width_min = 1,
    edge_width_max = 4,
    edge_color = None, # expect to range from 0 to 1
    edge_colormap = None,
    background_pos = None,
    background_ndcolor = "lavender",
    background_ndsize = 1,
    filename = "network_plot.pdf",
):
    """
    Plot cluster signaling network using matplotlib and NetworkX.
    
    Parameters
    ----------
    S : array-like
        Signaling matrix between clusters
    labels : list, optional
        Cluster labels
    node_size : float
        Size of nodes
    node_colormap : str
        Colormap for nodes
    node_cluster_colormap : dict, optional
        Custom color mapping for clusters
    node_pos : array-like, optional
        Positions for nodes
    edge_width_lb_quantile : float
        Lower bound quantile for edge width
    edge_width_ub_quantile : float
        Upper bound quantile for edge width
    edge_width_min : float
        Minimum edge width
    edge_width_max : float
        Maximum edge width
    edge_color : str or array-like
        Edge color specification
    edge_colormap : matplotlib colormap
        Colormap for edges when using numerical edge colors
    background_pos : array-like, optional
        Background positions to plot
    background_ndcolor : str
        Background color
    background_ndsize : float
        Background node size
    filename : str
        Output filename
    """
    
    if labels is None:
        labels = [str(i) for i in range(S.shape[0])]
    node_cmap = get_cmap_qualitative(node_colormap, n_colors=len(labels))
    
    # Create NetworkX graph
    G = nx.DiGraph()
    
    # Add nodes
    for i, label in enumerate(labels):
        G.add_node(label)
    
    # Add edges with weights
    edge_weights = []
    for i in range(S.shape[0]):
        for j in range(S.shape[1]):
            if S[i,j] > 0:
                G.add_edge(labels[i], labels[j], weight=S[i,j])
                edge_weights.append(S[i,j])
    
    # Set up the plot
    fig, ax = plt.subplots(figsize=(12, 10))
    
    # Calculate layout
    if node_pos is not None:
        pos = {labels[i]: node_pos[i,:2] for i in range(len(labels))}
    else:
        # Use spring layout as default
        pos = nx.spring_layout(G, k=2, iterations=50, seed=42)
    
    # Draw background positions if provided
    if background_pos is not None:
        ax.scatter(background_pos[:,0], background_pos[:,1], 
                   c=background_ndcolor, s=background_ndsize*50, alpha=0.3, zorder=1)
    
    # Calculate edge widths
    if edge_weights:
        edge_width_lb = np.quantile(edge_weights, edge_width_lb_quantile)
        edge_width_ub = np.quantile(edge_weights, edge_width_ub_quantile)
        normalized_weights = []
        for edge in G.edges(data=True):
            weight = edge[2]['weight']
            norm_weight = linear_clamp_value(weight, edge_width_lb, edge_width_ub, edge_width_min, edge_width_max)
            normalized_weights.append(norm_weight)
    else:
        normalized_weights = [1] * len(G.edges())
    
    # Calculate edge colors
    if isinstance(edge_color, str) and edge_color == "node":
        edge_colors = []
        for edge in G.edges():
            source_idx = labels.index(edge[0])
            edge_colors.append(node_cmap[source_idx % len(node_cmap)])
    elif isinstance(edge_color, np.ndarray):
        edge_colors = []
        for edge in G.edges():
            i, j = labels.index(edge[0]), labels.index(edge[1])
            color_val = edge_color[i, j] if edge_color.ndim == 2 else edge_color[i]
            if edge_colormap is not None:
                edge_colors.append(mpl.colors.to_hex(edge_colormap(color_val)))
            else:
                edge_colors.append(plt.cm.Greys(color_val))
    else:
        edge_colors = edge_color if edge_color else 'gray'
    
    # Draw edges
    if len(G.edges()) > 0:
        nx.draw_networkx_edges(G, pos, width=normalized_weights, 
                              edge_color=edge_colors, alpha=0.7, arrows=True, 
                              arrowsize=20, arrowstyle='->', 
                              connectionstyle='arc3,rad=0.1', ax=ax)
    
    # Calculate node colors
    node_colors = []
    for i, label in enumerate(labels):
        if node_cluster_colormap is None:
            node_colors.append(node_cmap[i % len(node_cmap)])
        else:
            node_colors.append(node_cluster_colormap[label])
    
    # Draw nodes
    nx.draw_networkx_nodes(G, pos, node_color=node_colors, 
                          node_size=node_size*8000, alpha=0.9, ax=ax)
    
    # Draw labels
    nx.draw_networkx_labels(G, pos, font_size=12, font_weight='bold', 
                           font_color='white', ax=ax)
    
    ax.set_aspect('equal')
    ax.axis('off')
    plt.tight_layout()
    
    # Save the plot
    if filename.endswith('.pdf'):
        plt.savefig(filename, format='pdf', bbox_inches='tight', dpi=300)
    elif filename.endswith('.png'):
        plt.savefig(filename, format='png', bbox_inches='tight', dpi=300)
    else:
        plt.savefig(filename + '.png', format='png', bbox_inches='tight', dpi=300)
    
    plt.close()

# Recommend plotting at most three LR species
def plot_cluster_signaling_network_multipair(S_list,
    p_values_list,
    labels = None,
    quantile_cutoff = None,
    p_value_cutoff = 0.05,
    node_size = 0.2,
    node_colormap = "Plotly",
    node_pos = None,
    edge_width_lb_quantile = 0.05,
    edge_width_ub_quantile = 0.95,
    edge_width_min = 1,
    edge_width_max = 4,
    edge_colors = ["blue","red","green","grey","yellow"],
    background_pos = None,
    background_ndcolor = "darkseagreen1",
    background_ndsize = 1,
    filename = "network_plot.pdf"
):
    """
    Plot cluster signaling network for multiple ligand-receptor pairs using matplotlib and NetworkX.
    
    Parameters
    ----------
    S_list : list
        List of signaling matrices
    p_values_list : list
        List of p-value matrices
    labels : list, optional
        Cluster labels
    quantile_cutoff : float, optional
        Quantile cutoff for significance
    p_value_cutoff : float
        P-value cutoff
    node_size : float
        Size of nodes
    node_colormap : str
        Colormap for nodes
    node_pos : array-like, optional
        Positions for nodes
    edge_width_lb_quantile : float
        Lower bound quantile for edge width
    edge_width_ub_quantile : float
        Upper bound quantile for edge width
    edge_width_min : float
        Minimum edge width
    edge_width_max : float
        Maximum edge width
    edge_colors : list
        Colors for different LR pairs
    background_pos : array-like, optional
        Background positions to plot
    background_ndcolor : str
        Background color
    background_ndsize : float
        Background node size
    filename : str
        Output filename
    """
    
    tmp_all_S = []
    for S in S_list:
        tmp_all_S.extend( list( S.reshape(-1) ) )
    tmp_all_S = np.array(tmp_all_S)
    if labels is None:
        labels = [str(i) for i in range(S_list[0].shape[0])]
    node_cmap = get_cmap_qualitative(node_colormap, n_colors=len(labels))
    
    # Create NetworkX graph
    G = nx.MultiDiGraph()
    
    # Add nodes
    for i, label in enumerate(labels):
        G.add_node(label)
    
    # Process each signaling matrix
    all_edge_weights = []
    all_edges_info = []  # Store edge information for each S matrix
    
    for s_idx, (S, p_values) in enumerate(zip(S_list, p_values_list)):
        edges_for_this_s = []
        
        for i in range(S.shape[0]):
            for j in range(S.shape[1]):
                include_edge = False
                
                if quantile_cutoff is not None:
                    cutoff = np.quantile(tmp_all_S, quantile_cutoff)
                    if S[i,j] >= cutoff:
                        include_edge = True
                
                if p_values[i,j] <= p_value_cutoff:
                    include_edge = True
                
                if include_edge:
                    edge_key = f"{s_idx}_{i}_{j}"  # Unique key for this edge
                    G.add_edge(labels[i], labels[j], key=edge_key, weight=S[i,j], s_idx=s_idx)
                    all_edge_weights.append(S[i,j])
                    edges_for_this_s.append((labels[i], labels[j], edge_key, S[i,j]))
        
        all_edges_info.append(edges_for_this_s)
    
    # Set up the plot
    fig, ax = plt.subplots(figsize=(12, 10))
    
    # Calculate layout
    if node_pos is not None:
        pos = {labels[i]: node_pos[i,:2] for i in range(len(labels))}
    else:
        # Use spring layout as default
        pos = nx.spring_layout(G, k=2, iterations=50, seed=42)
    
    # Draw background positions if provided
    if background_pos is not None:
        ax.scatter(background_pos[:,0], background_pos[:,1], 
                   c=background_ndcolor, s=background_ndsize*50, alpha=0.3, zorder=1)
    
    # Calculate edge width bounds
    if all_edge_weights:
        edge_width_lb = np.quantile(all_edge_weights, edge_width_lb_quantile)
        edge_width_ub = np.quantile(all_edge_weights, edge_width_ub_quantile)
    else:
        edge_width_lb, edge_width_ub = 0, 1
    
    # Draw edges for each signaling matrix with different colors
    for s_idx, edges_info in enumerate(all_edges_info):
        if not edges_info:
            continue
            
        # Prepare edges for this signaling matrix
        edges_to_draw = [(edge[0], edge[1]) for edge in edges_info]
        edge_weights = [edge[3] for edge in edges_info]
        
        # Normalize edge weights
        normalized_weights = [
            linear_clamp_value(w, edge_width_lb, edge_width_ub, edge_width_min, edge_width_max)
            for w in edge_weights
        ]
        
        # Get edge color for this signaling matrix
        edge_color = edge_colors[s_idx % len(edge_colors)]
        
        # Draw edges with slight offset to avoid overlap
        offset = s_idx * 0.05  # Small offset for visual separation
        pos_offset = {node: (x + offset, y + offset) for node, (x, y) in pos.items()}
        
        if edges_to_draw:
            nx.draw_networkx_edges(G, pos_offset, edgelist=edges_to_draw,
                                  width=normalized_weights, edge_color=edge_color, 
                                  alpha=0.7, arrows=True, arrowsize=15, 
                                  arrowstyle='->', connectionstyle=f'arc3,rad={0.1 + s_idx*0.05}', 
                                  ax=ax, label=f'Signal {s_idx+1}')
    
    # Calculate node colors
    node_colors = [node_cmap[i % len(node_cmap)] for i in range(len(labels))]
    
    # Draw nodes
    nx.draw_networkx_nodes(G, pos, node_color=node_colors, 
                          node_size=node_size*8000, alpha=0.9, ax=ax)
    
    # Draw labels
    nx.draw_networkx_labels(G, pos, font_size=12, font_weight='bold', 
                           font_color='white', ax=ax)
    
    ax.set_aspect('equal')
    ax.axis('off')
    plt.tight_layout()
    
    # Add legend for edge colors if multiple signaling matrices
    if len(S_list) > 1:
        legend_elements = [plt.Line2D([0], [0], color=edge_colors[i % len(edge_colors)], 
                                     lw=2, label=f'Signal {i+1}') 
                          for i in range(len(S_list))]
        ax.legend(handles=legend_elements, loc='upper right')
    
    # Save the plot
    if filename.endswith('.pdf'):
        plt.savefig(filename, format='pdf', bbox_inches='tight', dpi=300)
    elif filename.endswith('.png'):
        plt.savefig(filename, format='png', bbox_inches='tight', dpi=300)
    else:
        plt.savefig(filename + '.png', format='png', bbox_inches='tight', dpi=300)
    
    plt.close()

def plot_cell_signaling(X,
    V,
    signal_sum,
    cmap="coolwarm",
    cluster_cmap = None,
    arrow_color="tab:blue",
    plot_method="cell",
    background='summary',
    clustering=None,
    background_legend=False,
    adata=None,
    summary='sender',
    ndsize = 1,
    scale = 1.0,
    grid_density = 1,
    grid_knn = None,
    grid_scale = 1.0,
    grid_thresh = 1.0,
    grid_width = 0.005,
    stream_density = 1.0,
    stream_linewidth = 1,
    stream_cutoff_perc = 5,
    filename=None,
    ax = None,
    fig = None
):
    ndcolor = signal_sum
    ncell = X.shape[0]
    
    V_cell = V.copy()
    V_cell_sum = np.sum(V_cell, axis=1)
    V_cell[np.where(V_cell_sum==0)[0],:] = np.nan
    if summary == "sender":
        X_vec = X
    elif summary == "receiver":
        X_vec = X - V / scale

    if plot_method == "grid" or plot_method == "stream":
        # Get a rectangular grid
        xl, xr = np.min(X[:,0]), np.max(X[:,0])
        epsilon = 0.02*(xr-xl); xl -= epsilon; xr += epsilon
        yl, yr = np.min(X[:,1]), np.max(X[:,1])
        epsilon = 0.02*(yr-yl); yl -= epsilon; yr += epsilon
        ngrid_x = int(50 * grid_density)
        gridsize = (xr-xl) / float(ngrid_x)
        ngrid_y = int((yr-yl)/gridsize)
        meshgrid = np.meshgrid(np.linspace(xl,xr,ngrid_x), np.linspace(yl,yr,ngrid_y))
        grid_pts = np.concatenate((meshgrid[0].reshape(-1,1), meshgrid[1].reshape(-1,1)), axis=1)
    
        if grid_knn is None:
            grid_knn = int( X.shape[0] / 50 )
        nn_mdl = NearestNeighbors()
        nn_mdl.fit(X)
        dis, nbs = nn_mdl.kneighbors(grid_pts, n_neighbors=grid_knn)
        w = norm.pdf(x=dis, scale=gridsize * grid_scale)
        w_sum = w.sum(axis=1)

        V_grid = (V[nbs] * w[:,:,None]).sum(axis=1)
        V_grid /= np.maximum(1, w_sum)[:,None]

        if plot_method == "grid":
            grid_thresh *= np.percentile(w_sum, 99) / 100
            grid_pts, V_grid = grid_pts[w_sum > grid_thresh], V_grid[w_sum > grid_thresh]
        elif plot_method == "stream":
            x_grid = np.linspace(xl, xr, ngrid_x)
            y_grid = np.linspace(yl, yr, ngrid_y)
            V_grid = V_grid.T.reshape(2, ngrid_y, ngrid_x)
            vlen = np.sqrt((V_grid ** 2).sum(0))
            grid_thresh = 10 ** (grid_thresh - 6)
            grid_thresh = np.clip(grid_thresh, None, np.max(vlen) * 0.9)
            cutoff = vlen.reshape(V_grid[0].shape) < grid_thresh
            length = np.sum(np.mean(np.abs(V[nbs]),axis=1),axis=1).T
            length = length.reshape(ngrid_y, ngrid_x)
            cutoff |= length < np.percentile(length, stream_cutoff_perc)

            V_grid[0][cutoff] = np.nan

    if cmap == 'Plotly':
        cmap = plotly.colors.qualitative.Plotly
    elif cmap == 'Light24':
        cmap = plotly.colors.qualitative.Light24
    elif cmap == 'Dark24':
        cmap = plotly.colors.qualitative.Dark24
    elif cmap == 'Alphabet':
        cmap = plotly.colors.qualitative.Alphabet

    idx = np.argsort(ndcolor)
    if background == 'summary' or background == 'cluster':
        if not ndsize==0:
            if background == 'summary':
                ax.scatter(X[idx,0], X[idx,1], s=ndsize, c=ndcolor[idx], cmap=cmap, linewidth=0)
            elif background == 'cluster':
                adata.obs[clustering]=adata.obs[clustering].astype('category')
                labels = np.array( adata.obs[clustering], str )
                unique_labels = adata.obs[clustering].cat.categories
                for i_label in range(len(unique_labels)):
                    idx = np.where(labels == unique_labels[i_label])[0]
                    if cluster_cmap is None:
                        ax.scatter(X[idx,0], X[idx,1], s=ndsize, c=cmap[i_label], linewidth=0, label=unique_labels[i_label])
                    elif not cluster_cmap is None:
                        ax.scatter(X[idx,0], X[idx,1], s=ndsize, c=cluster_cmap[unique_labels[i_label]], linewidth=0, label=unique_labels[i_label])
                if background_legend:
                    ax.legend(markerscale=2.0, loc=[1.0,0.0])
        if plot_method == "cell":
            ax.quiver(X_vec[:,0], X_vec[:,1], V_cell[:,0], V_cell[:,1], scale=scale, scale_units='x', color=arrow_color)
        elif plot_method == "grid":
            ax.quiver(grid_pts[:,0], grid_pts[:,1], V_grid[:,0], V_grid[:,1], scale=scale, scale_units='x', width=grid_width, color=arrow_color)
        elif plot_method == "stream":
            lengths = np.sqrt((V_grid ** 2).sum(0))
            stream_linewidth *= 2 * lengths / lengths[~np.isnan(lengths)].max()
            ax.streamplot(x_grid, y_grid, V_grid[0], V_grid[1], color=arrow_color, density=stream_density, linewidth=stream_linewidth)
    elif background == 'image':
        spatial_mapping = adata.uns.get("spatial", {})
        library_id = list(spatial_mapping.keys())[0]
        spatial_data = spatial_mapping[library_id]
        img = spatial_data['images']['hires']
        sf = spatial_data['scalefactors']['tissue_hires_scalef']
        ax.imshow(img, origin='lower')
        if plot_method == "cell":
            ax.quiver(X_vec[:,0]*sf, X_vec[:,1]*sf, V_cell[:,0]*sf, V_cell[:,1]*sf, scale=scale, scale_units='x', color=arrow_color)
        elif plot_method == "grid":
            ax.quiver(grid_pts[:,0]*sf, grid_pts[:,1]*sf, V_grid[:,0]*sf, V_grid[:,1]*sf, scale=scale, scale_units='x', width=grid_width, color=arrow_color)
        elif plot_method == "stream":
            lengths = np.sqrt((V_grid ** 2).sum(0))
            stream_linewidth *= 2 * lengths / lengths[~np.isnan(lengths)].max()
            ax.streamplot(x_grid*sf, y_grid*sf, V_grid[0]*sf, V_grid[1]*sf, color=arrow_color, density=stream_density, linewidth=stream_linewidth)

    ax.axis("equal")
    ax.axis("off")
    if not filename is None:
        plt.savefig(filename, dpi=500, bbox_inches = 'tight', transparent=True)


def plot_cell_signaling_compare(X,
    V,
    ax,
    arrow_color="tab:blue",
    plot_method="cell",
    summary='sender',
    scale = 1.0,
    grid_density = 1,
    grid_knn = None,
    grid_scale = 1.0,
    grid_thresh = 1.0,
):
    V_cell = V.copy()
    V_cell_sum = np.sum(V_cell, axis=1)
    V_cell[np.where(V_cell_sum==0)[0],:] = np.nan
    if summary == "sender":
        X_vec = X
    elif summary == "receiver":
        X_vec = X - V / scale

    if plot_method == "grid":
        # Get a rectangular grid
        xl, xr = np.min(X[:,0]), np.max(X[:,0])
        epsilon = 0.02*(xr-xl); xl -= epsilon; xr += epsilon
        yl, yr = np.min(X[:,1]), np.max(X[:,1])
        epsilon = 0.02*(yr-yl); yl -= epsilon; yr += epsilon
        ngrid_x = int(50 * grid_density)
        gridsize = (xr-xl) / float(ngrid_x)
        ngrid_y = int((yr-yl)/gridsize)
        meshgrid = np.meshgrid(np.linspace(xl,xr,ngrid_x), np.linspace(yl,yr,ngrid_y))
        grid_pts = np.concatenate((meshgrid[0].reshape(-1,1), meshgrid[1].reshape(-1,1)), axis=1)
    
        if grid_knn is None:
            grid_knn = int( X.shape[0] / 50 )
        nn_mdl = NearestNeighbors()
        nn_mdl.fit(X)
        dis, nbs = nn_mdl.kneighbors(grid_pts, n_neighbors=grid_knn)
        w = norm.pdf(x=dis, scale=gridsize * grid_scale)
        w_sum = w.sum(axis=1)

        V_grid = (V[nbs] * w[:,:,None]).sum(axis=1)
        V_grid /= np.maximum(1, w_sum)[:,None]

        if plot_method == "grid":
            grid_thresh *= np.percentile(w_sum, 99) / 100
            grid_pts, V_grid = grid_pts[w_sum > grid_thresh], V_grid[w_sum > grid_thresh]

    if plot_method == "cell":
        ax.quiver(X_vec[:,0], X_vec[:,1], V_cell[:,0], V_cell[:,1], scale=scale, scale_units='x', color=arrow_color, alpha=1.0)
    elif plot_method == "grid":
        ax.quiver(grid_pts[:,0], grid_pts[:,1], V_grid[:,0], V_grid[:,1], scale=scale, scale_units='x', color=arrow_color, alpha=1.0)
    return ax


from typing import Union, List, Optional, Iterable, Sequence, Dict
import warnings

from matplotlib.axes import Axes
import numpy as np
import pandas as pd
from scipy.sparse import issparse
import matplotlib.pyplot as plt
import seaborn as sns
import scanpy as sc
from anndata import AnnData
from ._single import embedding

def embedding_multi(
    data: AnnData,
    basis: str,
    color: Optional[Union[str, Sequence[str]]] = None,
    use_raw: Optional[bool] = None,
    layer: Optional[str] = None,
    **kwargs,
):
    r"""
    Create embedding scatter plots for multi-modal data (MuData) or single-cell data.
    
    Produces scatter plots on specified embeddings, supporting cross-modality feature visualization.
    For modality-specific embeddings, use format 'modality:embedding' (e.g., 'rna:X_pca').
    
    Arguments:
        data: AnnData or MuData object containing embedding and feature data
        basis: Name of embedding in obsm (e.g., 'X_umap') or modality-specific ('rna:X_pca')
        color: Gene names or obs columns to color points by (None)
        use_raw: Whether to use .raw attribute for features (None, auto-determined)
        layer: Specific data layer to use for coloring (None)
        **kwargs: Additional arguments passed to embedding plotting function
        
    Returns:
        Plot axes or figure depending on underlying plotting function
    """
    if isinstance(data, AnnData):
        return embedding(
            data, basis=basis, color=color, use_raw=use_raw, layer=layer, **kwargs
        )

    # `data` is MuData
    if basis not in data.obsm and "X_" + basis in data.obsm:
        basis = "X_" + basis

    if basis in data.obsm:
        adata = data
        basis_mod = basis
    else:
        # basis is not a joint embedding
        try:
            mod, basis_mod = basis.split(":")
        except ValueError:
            raise ValueError(f"Basis {basis} is not present in the MuData object (.obsm)")

        if mod not in data.mod:
            raise ValueError(
                f"Modality {mod} is not present in the MuData object with modalities {', '.join(data.mod)}"
            )

        adata = data.mod[mod]
        if basis_mod not in adata.obsm:
            if "X_" + basis_mod in adata.obsm:
                basis_mod = "X_" + basis_mod
            elif len(adata.obsm) > 0:
                raise ValueError(
                    f"Basis {basis_mod} is not present in the modality {mod} with embeddings {', '.join(adata.obsm)}"
                )
            else:
                raise ValueError(
                    f"Basis {basis_mod} is not present in the modality {mod} with no embeddings"
                )

    obs = data.obs.loc[adata.obs.index.values]

    if color is None:
        ad = AnnData(obs=obs, obsm=adata.obsm, obsp=adata.obsp)
        return sc.pl.embedding(ad, basis=basis_mod, **kwargs)

    # Some `color` has been provided
    if isinstance(color, str):
        keys = color = [color]
    elif isinstance(color, Iterable):
        keys = color
    else:
        raise TypeError("Expected color to be a string or an iterable.")

    # Fetch respective features
    if not all([key in obs for key in keys]):
        # {'rna': [True, False], 'prot': [False, True]}
        keys_in_mod = {m: [key in data.mod[m].var_names for key in keys] for m in data.mod}

        # .raw slots might have exclusive var_names
        if use_raw is None or use_raw:
            for i, k in enumerate(keys):
                for m in data.mod:
                    if keys_in_mod[m][i] == False and data.mod[m].raw is not None:
                        keys_in_mod[m][i] = k in data.mod[m].raw.var_names

        # e.g. color="rna:CD8A" - especially relevant for mdata.axis == -1
        mod_key_modifier: dict[str, str] = dict()
        for i, k in enumerate(keys):
            mod_key_modifier[k] = k
            for m in data.mod:
                if not keys_in_mod[m][i]:
                    k_clean = k
                    if k.startswith(f"{m}:"):
                        k_clean = k.split(":", 1)[1]

                    keys_in_mod[m][i] = k_clean in data.mod[m].var_names
                    if keys_in_mod[m][i]:
                        mod_key_modifier[k] = k_clean
                    if use_raw is None or use_raw:
                        if keys_in_mod[m][i] == False and data.mod[m].raw is not None:
                            keys_in_mod[m][i] = k_clean in data.mod[m].raw.var_names

        for m in data.mod:
            if np.sum(keys_in_mod[m]) > 0:
                mod_keys = np.array(keys)[keys_in_mod[m]]
                mod_keys = np.array([mod_key_modifier[k] for k in mod_keys])

                if use_raw is None or use_raw:
                    if data.mod[m].raw is not None:
                        keysidx = data.mod[m].raw.var.index.get_indexer_for(mod_keys)
                        fmod_adata = AnnData(
                            X=data.mod[m].raw.X[:, keysidx],
                            var=pd.DataFrame(index=mod_keys),
                            obs=data.mod[m].obs,
                        )
                    else:
                        if use_raw:
                            warnings.warn(
                                f"Attibute .raw is None for the modality {m}, using .X instead"
                            )
                        fmod_adata = data.mod[m][:, mod_keys]
                else:
                    fmod_adata = data.mod[m][:, mod_keys]

                if layer is not None:
                    if isinstance(layer, Dict):
                        m_layer = layer.get(m, None)
                        if m_layer is not None:
                            x = data.mod[m][:, mod_keys].layers[m_layer]
                            fmod_adata.X = x.todense() if issparse(x) else x
                            if use_raw:
                                warnings.warn(f"Layer='{layer}' superseded use_raw={use_raw}")
                    elif layer in data.mod[m].layers:
                        x = data.mod[m][:, mod_keys].layers[layer]
                        fmod_adata.X = x.todense() if issparse(x) else x
                        if use_raw:
                            warnings.warn(f"Layer='{layer}' superseded use_raw={use_raw}")
                    else:
                        warnings.warn(
                            f"Layer {layer} is not present for the modality {m}, using count matrix instead"
                        )
                x = fmod_adata.X.toarray() if issparse(fmod_adata.X) else fmod_adata.X
                obs = obs.join(
                    pd.DataFrame(x, columns=mod_keys, index=fmod_adata.obs_names),
                    how="left",
                )

        color = [mod_key_modifier[k] for k in keys]

    ad = AnnData(obs=obs, obsm=adata.obsm, obsp=adata.obsp, uns=adata.uns)
    retval = embedding(ad, basis=basis_mod, color=color, **kwargs)
    for key, col in zip(keys, color):
        try:
            adata.uns[f"{key}_colors"] = ad.uns[f"{col}_colors"]
        except KeyError:
            pass
    return retval